from google.protobuf import timestamp_pb2 as _timestamp_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class KeyphrasesBoosting(_message.Message):
    __slots__ = ("keyphrases_boosting",)
    KEYPHRASES_BOOSTING_FIELD_NUMBER: _ClassVar[int]
    keyphrases_boosting: _containers.RepeatedCompositeFieldContainer[KeyphraseBoosting]
    def __init__(self, keyphrases_boosting: _Optional[_Iterable[_Union[KeyphraseBoosting, _Mapping]]] = ...) -> None: ...

class KeyphraseBoosting(_message.Message):
    __slots__ = ("id", "keyphrase", "level", "created_at", "created_by", "updated_at", "updated_by")
    ID_FIELD_NUMBER: _ClassVar[int]
    KEYPHRASE_FIELD_NUMBER: _ClassVar[int]
    LEVEL_FIELD_NUMBER: _ClassVar[int]
    CREATED_AT_FIELD_NUMBER: _ClassVar[int]
    CREATED_BY_FIELD_NUMBER: _ClassVar[int]
    UPDATED_AT_FIELD_NUMBER: _ClassVar[int]
    UPDATED_BY_FIELD_NUMBER: _ClassVar[int]
    id: str
    keyphrase: str
    level: str
    created_at: _timestamp_pb2.Timestamp
    created_by: str
    updated_at: _timestamp_pb2.Timestamp
    updated_by: str
    def __init__(self, id: _Optional[str] = ..., keyphrase: _Optional[str] = ..., level: _Optional[str] = ..., created_at: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., created_by: _Optional[str] = ..., updated_at: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., updated_by: _Optional[str] = ...) -> None: ...

class KeyphraseBoosting_CreateKeyphrase(_message.Message):
    __slots__ = ("id", "keyphrase", "level")
    ID_FIELD_NUMBER: _ClassVar[int]
    KEYPHRASE_FIELD_NUMBER: _ClassVar[int]
    LEVEL_FIELD_NUMBER: _ClassVar[int]
    id: str
    keyphrase: str
    level: str
    def __init__(self, id: _Optional[str] = ..., keyphrase: _Optional[str] = ..., level: _Optional[str] = ...) -> None: ...

class KeyphraseBoosting_UpdateKeyphrase(_message.Message):
    __slots__ = ("id", "keyphrase", "level")
    ID_FIELD_NUMBER: _ClassVar[int]
    KEYPHRASE_FIELD_NUMBER: _ClassVar[int]
    LEVEL_FIELD_NUMBER: _ClassVar[int]
    id: str
    keyphrase: str
    level: str
    def __init__(self, id: _Optional[str] = ..., keyphrase: _Optional[str] = ..., level: _Optional[str] = ...) -> None: ...

class KeyphraseBoosting_DeleteKeyphrase(_message.Message):
    __slots__ = ("id",)
    ID_FIELD_NUMBER: _ClassVar[int]
    id: str
    def __init__(self, id: _Optional[str] = ...) -> None: ...



from google.protobuf import timestamp_pb2 as _timestamp_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class FunctionParameter(_message.Message):
    __slots__ = ("id", "name", "description", "type", "created_at", "created_by", "updated_at", "updated_by")
    ID_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    DESCRIPTION_FIELD_NUMBER: _ClassVar[int]
    TYPE_FIELD_NUMBER: _ClassVar[int]
    CREATED_AT_FIELD_NUMBER: _ClassVar[int]
    CREATED_BY_FIELD_NUMBER: _ClassVar[int]
    UPDATED_AT_FIELD_NUMBER: _ClassVar[int]
    UPDATED_BY_FIELD_NUMBER: _ClassVar[int]
    id: str
    name: str
    description: str
    type: str
    created_at: _timestamp_pb2.Timestamp
    created_by: str
    updated_at: _timestamp_pb2.Timestamp
    updated_by: str
    def __init__(self, id: _Optional[str] = ..., name: _Optional[str] = ..., description: _Optional[str] = ..., type: _Optional[str] = ..., created_at: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., created_by: _Optional[str] = ..., updated_at: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., updated_by: _Optional[str] = ...) -> None: ...

class FunctionError(_message.Message):
    __slots__ = ("lineno", "message", "text")
    LINENO_FIELD_NUMBER: _ClassVar[int]
    MESSAGE_FIELD_NUMBER: _ClassVar[int]
    TEXT_FIELD_NUMBER: _ClassVar[int]
    lineno: int
    message: str
    text: str
    def __init__(self, lineno: _Optional[int] = ..., message: _Optional[str] = ..., text: _Optional[str] = ...) -> None: ...

class FunctionDelayResponseReferences(_message.Message):
    __slots__ = ("variables",)
    class VariablesEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: bool
        def __init__(self, key: _Optional[str] = ..., value: bool = ...) -> None: ...
    VARIABLES_FIELD_NUMBER: _ClassVar[int]
    variables: _containers.ScalarMap[str, bool]
    def __init__(self, variables: _Optional[_Mapping[str, bool]] = ...) -> None: ...

class FunctionDelayResponse(_message.Message):
    __slots__ = ("id", "message", "duration", "created_at", "created_by", "updated_at", "updated_by", "references")
    ID_FIELD_NUMBER: _ClassVar[int]
    MESSAGE_FIELD_NUMBER: _ClassVar[int]
    DURATION_FIELD_NUMBER: _ClassVar[int]
    CREATED_AT_FIELD_NUMBER: _ClassVar[int]
    CREATED_BY_FIELD_NUMBER: _ClassVar[int]
    UPDATED_AT_FIELD_NUMBER: _ClassVar[int]
    UPDATED_BY_FIELD_NUMBER: _ClassVar[int]
    REFERENCES_FIELD_NUMBER: _ClassVar[int]
    id: str
    message: str
    duration: int
    created_at: _timestamp_pb2.Timestamp
    created_by: str
    updated_at: _timestamp_pb2.Timestamp
    updated_by: str
    references: FunctionDelayResponseReferences
    def __init__(self, id: _Optional[str] = ..., message: _Optional[str] = ..., duration: _Optional[int] = ..., created_at: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., created_by: _Optional[str] = ..., updated_at: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., updated_by: _Optional[str] = ..., references: _Optional[_Union[FunctionDelayResponseReferences, _Mapping]] = ...) -> None: ...

class FunctionLatencyControl(_message.Message):
    __slots__ = ("enabled", "delay_responses", "initial_delay", "interval", "created_at", "created_by", "updated_at", "updated_by")
    ENABLED_FIELD_NUMBER: _ClassVar[int]
    DELAY_RESPONSES_FIELD_NUMBER: _ClassVar[int]
    INITIAL_DELAY_FIELD_NUMBER: _ClassVar[int]
    INTERVAL_FIELD_NUMBER: _ClassVar[int]
    CREATED_AT_FIELD_NUMBER: _ClassVar[int]
    CREATED_BY_FIELD_NUMBER: _ClassVar[int]
    UPDATED_AT_FIELD_NUMBER: _ClassVar[int]
    UPDATED_BY_FIELD_NUMBER: _ClassVar[int]
    enabled: bool
    delay_responses: _containers.RepeatedCompositeFieldContainer[FunctionDelayResponse]
    initial_delay: int
    interval: int
    created_at: _timestamp_pb2.Timestamp
    created_by: str
    updated_at: _timestamp_pb2.Timestamp
    updated_by: str
    def __init__(self, enabled: bool = ..., delay_responses: _Optional[_Iterable[_Union[FunctionDelayResponse, _Mapping]]] = ..., initial_delay: _Optional[int] = ..., interval: _Optional[int] = ..., created_at: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., created_by: _Optional[str] = ..., updated_at: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., updated_by: _Optional[str] = ...) -> None: ...

class FunctionReferences(_message.Message):
    __slots__ = ("flow_steps", "topics", "stop_keywords", "behaviour", "variables")
    class FlowStepsEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: bool
        def __init__(self, key: _Optional[str] = ..., value: bool = ...) -> None: ...
    class TopicsEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: bool
        def __init__(self, key: _Optional[str] = ..., value: bool = ...) -> None: ...
    class StopKeywordsEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: bool
        def __init__(self, key: _Optional[str] = ..., value: bool = ...) -> None: ...
    class BehaviourEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: bool
        def __init__(self, key: _Optional[str] = ..., value: bool = ...) -> None: ...
    class VariablesEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: bool
        def __init__(self, key: _Optional[str] = ..., value: bool = ...) -> None: ...
    FLOW_STEPS_FIELD_NUMBER: _ClassVar[int]
    TOPICS_FIELD_NUMBER: _ClassVar[int]
    STOP_KEYWORDS_FIELD_NUMBER: _ClassVar[int]
    BEHAVIOUR_FIELD_NUMBER: _ClassVar[int]
    VARIABLES_FIELD_NUMBER: _ClassVar[int]
    flow_steps: _containers.ScalarMap[str, bool]
    topics: _containers.ScalarMap[str, bool]
    stop_keywords: _containers.ScalarMap[str, bool]
    behaviour: _containers.ScalarMap[str, bool]
    variables: _containers.ScalarMap[str, bool]
    def __init__(self, flow_steps: _Optional[_Mapping[str, bool]] = ..., topics: _Optional[_Mapping[str, bool]] = ..., stop_keywords: _Optional[_Mapping[str, bool]] = ..., behaviour: _Optional[_Mapping[str, bool]] = ..., variables: _Optional[_Mapping[str, bool]] = ...) -> None: ...

class Function(_message.Message):
    __slots__ = ("id", "name", "description", "parameters", "code", "errors", "latency_control", "created_at", "created_by", "updated_at", "updated_by", "archived", "references")
    ID_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    DESCRIPTION_FIELD_NUMBER: _ClassVar[int]
    PARAMETERS_FIELD_NUMBER: _ClassVar[int]
    CODE_FIELD_NUMBER: _ClassVar[int]
    ERRORS_FIELD_NUMBER: _ClassVar[int]
    LATENCY_CONTROL_FIELD_NUMBER: _ClassVar[int]
    CREATED_AT_FIELD_NUMBER: _ClassVar[int]
    CREATED_BY_FIELD_NUMBER: _ClassVar[int]
    UPDATED_AT_FIELD_NUMBER: _ClassVar[int]
    UPDATED_BY_FIELD_NUMBER: _ClassVar[int]
    ARCHIVED_FIELD_NUMBER: _ClassVar[int]
    REFERENCES_FIELD_NUMBER: _ClassVar[int]
    id: str
    name: str
    description: str
    parameters: _containers.RepeatedCompositeFieldContainer[FunctionParameter]
    code: str
    errors: _containers.RepeatedCompositeFieldContainer[FunctionError]
    latency_control: FunctionLatencyControl
    created_at: _timestamp_pb2.Timestamp
    created_by: str
    updated_at: _timestamp_pb2.Timestamp
    updated_by: str
    archived: bool
    references: FunctionReferences
    def __init__(self, id: _Optional[str] = ..., name: _Optional[str] = ..., description: _Optional[str] = ..., parameters: _Optional[_Iterable[_Union[FunctionParameter, _Mapping]]] = ..., code: _Optional[str] = ..., errors: _Optional[_Iterable[_Union[FunctionError, _Mapping]]] = ..., latency_control: _Optional[_Union[FunctionLatencyControl, _Mapping]] = ..., created_at: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., created_by: _Optional[str] = ..., updated_at: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., updated_by: _Optional[str] = ..., archived: bool = ..., references: _Optional[_Union[FunctionReferences, _Mapping]] = ...) -> None: ...

class Functions(_message.Message):
    __slots__ = ("functions",)
    FUNCTIONS_FIELD_NUMBER: _ClassVar[int]
    functions: _containers.RepeatedCompositeFieldContainer[Function]
    def __init__(self, functions: _Optional[_Iterable[_Union[Function, _Mapping]]] = ...) -> None: ...

class FunctionParameterUpdate(_message.Message):
    __slots__ = ("id", "name", "description", "type")
    ID_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    DESCRIPTION_FIELD_NUMBER: _ClassVar[int]
    TYPE_FIELD_NUMBER: _ClassVar[int]
    id: str
    name: str
    description: str
    type: str
    def __init__(self, id: _Optional[str] = ..., name: _Optional[str] = ..., description: _Optional[str] = ..., type: _Optional[str] = ...) -> None: ...

class FunctionCreateLatencyControl(_message.Message):
    __slots__ = ("enabled", "delay_responses", "initial_delay", "interval")
    ENABLED_FIELD_NUMBER: _ClassVar[int]
    DELAY_RESPONSES_FIELD_NUMBER: _ClassVar[int]
    INITIAL_DELAY_FIELD_NUMBER: _ClassVar[int]
    INTERVAL_FIELD_NUMBER: _ClassVar[int]
    enabled: bool
    delay_responses: _containers.RepeatedCompositeFieldContainer[FunctionDelayResponse]
    initial_delay: int
    interval: int
    def __init__(self, enabled: bool = ..., delay_responses: _Optional[_Iterable[_Union[FunctionDelayResponse, _Mapping]]] = ..., initial_delay: _Optional[int] = ..., interval: _Optional[int] = ...) -> None: ...

class Function_CreateFunction(_message.Message):
    __slots__ = ("id", "name", "description", "parameters", "code", "errors", "latency_control", "references", "archived")
    ID_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    DESCRIPTION_FIELD_NUMBER: _ClassVar[int]
    PARAMETERS_FIELD_NUMBER: _ClassVar[int]
    CODE_FIELD_NUMBER: _ClassVar[int]
    ERRORS_FIELD_NUMBER: _ClassVar[int]
    LATENCY_CONTROL_FIELD_NUMBER: _ClassVar[int]
    REFERENCES_FIELD_NUMBER: _ClassVar[int]
    ARCHIVED_FIELD_NUMBER: _ClassVar[int]
    id: str
    name: str
    description: str
    parameters: _containers.RepeatedCompositeFieldContainer[FunctionParameterUpdate]
    code: str
    errors: _containers.RepeatedCompositeFieldContainer[FunctionError]
    latency_control: FunctionCreateLatencyControl
    references: FunctionReferences
    archived: bool
    def __init__(self, id: _Optional[str] = ..., name: _Optional[str] = ..., description: _Optional[str] = ..., parameters: _Optional[_Iterable[_Union[FunctionParameterUpdate, _Mapping]]] = ..., code: _Optional[str] = ..., errors: _Optional[_Iterable[_Union[FunctionError, _Mapping]]] = ..., latency_control: _Optional[_Union[FunctionCreateLatencyControl, _Mapping]] = ..., references: _Optional[_Union[FunctionReferences, _Mapping]] = ..., archived: bool = ...) -> None: ...

class ParametersUpdate(_message.Message):
    __slots__ = ("parameters",)
    PARAMETERS_FIELD_NUMBER: _ClassVar[int]
    parameters: _containers.RepeatedCompositeFieldContainer[FunctionParameterUpdate]
    def __init__(self, parameters: _Optional[_Iterable[_Union[FunctionParameterUpdate, _Mapping]]] = ...) -> None: ...

class ErrorsUpdate(_message.Message):
    __slots__ = ("errors",)
    ERRORS_FIELD_NUMBER: _ClassVar[int]
    errors: _containers.RepeatedCompositeFieldContainer[FunctionError]
    def __init__(self, errors: _Optional[_Iterable[_Union[FunctionError, _Mapping]]] = ...) -> None: ...

class Function_UpdateFunction(_message.Message):
    __slots__ = ("id", "name", "description", "parameters", "code", "errors", "references", "archived")
    ID_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    DESCRIPTION_FIELD_NUMBER: _ClassVar[int]
    PARAMETERS_FIELD_NUMBER: _ClassVar[int]
    CODE_FIELD_NUMBER: _ClassVar[int]
    ERRORS_FIELD_NUMBER: _ClassVar[int]
    REFERENCES_FIELD_NUMBER: _ClassVar[int]
    ARCHIVED_FIELD_NUMBER: _ClassVar[int]
    id: str
    name: str
    description: str
    parameters: ParametersUpdate
    code: str
    errors: ErrorsUpdate
    references: FunctionReferences
    archived: bool
    def __init__(self, id: _Optional[str] = ..., name: _Optional[str] = ..., description: _Optional[str] = ..., parameters: _Optional[_Union[ParametersUpdate, _Mapping]] = ..., code: _Optional[str] = ..., errors: _Optional[_Union[ErrorsUpdate, _Mapping]] = ..., references: _Optional[_Union[FunctionReferences, _Mapping]] = ..., archived: bool = ...) -> None: ...

class Function_DeleteFunction(_message.Message):
    __slots__ = ("id",)
    ID_FIELD_NUMBER: _ClassVar[int]
    id: str
    def __init__(self, id: _Optional[str] = ...) -> None: ...

class DelayResponseUpdate(_message.Message):
    __slots__ = ("id", "message", "duration", "references")
    ID_FIELD_NUMBER: _ClassVar[int]
    MESSAGE_FIELD_NUMBER: _ClassVar[int]
    DURATION_FIELD_NUMBER: _ClassVar[int]
    REFERENCES_FIELD_NUMBER: _ClassVar[int]
    id: str
    message: str
    duration: int
    references: FunctionDelayResponseReferences
    def __init__(self, id: _Optional[str] = ..., message: _Optional[str] = ..., duration: _Optional[int] = ..., references: _Optional[_Union[FunctionDelayResponseReferences, _Mapping]] = ...) -> None: ...

class DelayResponsesUpdate(_message.Message):
    __slots__ = ("delay_responses",)
    DELAY_RESPONSES_FIELD_NUMBER: _ClassVar[int]
    delay_responses: _containers.RepeatedCompositeFieldContainer[DelayResponseUpdate]
    def __init__(self, delay_responses: _Optional[_Iterable[_Union[DelayResponseUpdate, _Mapping]]] = ...) -> None: ...

class Function_UpdateLatencyControl(_message.Message):
    __slots__ = ("function_id", "enabled", "delay_responses", "initial_delay", "interval")
    FUNCTION_ID_FIELD_NUMBER: _ClassVar[int]
    ENABLED_FIELD_NUMBER: _ClassVar[int]
    DELAY_RESPONSES_FIELD_NUMBER: _ClassVar[int]
    INITIAL_DELAY_FIELD_NUMBER: _ClassVar[int]
    INTERVAL_FIELD_NUMBER: _ClassVar[int]
    function_id: str
    enabled: bool
    delay_responses: DelayResponsesUpdate
    initial_delay: int
    interval: int
    def __init__(self, function_id: _Optional[str] = ..., enabled: bool = ..., delay_responses: _Optional[_Union[DelayResponsesUpdate, _Mapping]] = ..., initial_delay: _Optional[int] = ..., interval: _Optional[int] = ...) -> None: ...



from google.protobuf import timestamp_pb2 as _timestamp_pb2
from poly.handlers.protobuf import llm_settings_pb2 as _llm_settings_pb2
from poly.handlers.protobuf import content_filter_settings_pb2 as _content_filter_settings_pb2
from poly.handlers.protobuf import asr_settings_pb2 as _asr_settings_pb2
from poly.handlers.protobuf import agent_settings_pb2 as _agent_settings_pb2
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class ChannelType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    VOICE: _ClassVar[ChannelType]
    WEB_CHAT: _ClassVar[ChannelType]

class ChannelStatus(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    NOT_CREATED: _ClassVar[ChannelStatus]
    CREATED: _ClassVar[ChannelStatus]
VOICE: ChannelType
WEB_CHAT: ChannelType
NOT_CREATED: ChannelStatus
CREATED: ChannelStatus

class StylePrompt(_message.Message):
    __slots__ = ("prompt", "created_at", "created_by", "updated_at", "updated_by")
    PROMPT_FIELD_NUMBER: _ClassVar[int]
    CREATED_AT_FIELD_NUMBER: _ClassVar[int]
    CREATED_BY_FIELD_NUMBER: _ClassVar[int]
    UPDATED_AT_FIELD_NUMBER: _ClassVar[int]
    UPDATED_BY_FIELD_NUMBER: _ClassVar[int]
    prompt: str
    created_at: str
    created_by: str
    updated_at: _timestamp_pb2.Timestamp
    updated_by: str
    def __init__(self, prompt: _Optional[str] = ..., created_at: _Optional[str] = ..., created_by: _Optional[str] = ..., updated_at: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., updated_by: _Optional[str] = ...) -> None: ...

class StylePrompt_UpdateStylePrompt(_message.Message):
    __slots__ = ("prompt",)
    PROMPT_FIELD_NUMBER: _ClassVar[int]
    prompt: str
    def __init__(self, prompt: _Optional[str] = ...) -> None: ...

class ChannelConfig(_message.Message):
    __slots__ = ("llm_settings", "style_prompt", "greeting", "safety_filters")
    LLM_SETTINGS_FIELD_NUMBER: _ClassVar[int]
    STYLE_PROMPT_FIELD_NUMBER: _ClassVar[int]
    GREETING_FIELD_NUMBER: _ClassVar[int]
    SAFETY_FILTERS_FIELD_NUMBER: _ClassVar[int]
    llm_settings: _llm_settings_pb2.LLMSettings
    style_prompt: StylePrompt
    greeting: _agent_settings_pb2.Greeting
    safety_filters: _content_filter_settings_pb2.ContentFilterSettings
    def __init__(self, llm_settings: _Optional[_Union[_llm_settings_pb2.LLMSettings, _Mapping]] = ..., style_prompt: _Optional[_Union[StylePrompt, _Mapping]] = ..., greeting: _Optional[_Union[_agent_settings_pb2.Greeting, _Mapping]] = ..., safety_filters: _Optional[_Union[_content_filter_settings_pb2.ContentFilterSettings, _Mapping]] = ...) -> None: ...

class VoiceChannel(_message.Message):
    __slots__ = ("config", "asr_settings", "disclaimer")
    CONFIG_FIELD_NUMBER: _ClassVar[int]
    ASR_SETTINGS_FIELD_NUMBER: _ClassVar[int]
    DISCLAIMER_FIELD_NUMBER: _ClassVar[int]
    config: ChannelConfig
    asr_settings: _asr_settings_pb2.ASRSettings
    disclaimer: _agent_settings_pb2.DisclaimerMessage
    def __init__(self, config: _Optional[_Union[ChannelConfig, _Mapping]] = ..., asr_settings: _Optional[_Union[_asr_settings_pb2.ASRSettings, _Mapping]] = ..., disclaimer: _Optional[_Union[_agent_settings_pb2.DisclaimerMessage, _Mapping]] = ...) -> None: ...

class WebChatChannel(_message.Message):
    __slots__ = ("config", "status")
    CONFIG_FIELD_NUMBER: _ClassVar[int]
    STATUS_FIELD_NUMBER: _ClassVar[int]
    config: ChannelConfig
    status: ChannelStatus
    def __init__(self, config: _Optional[_Union[ChannelConfig, _Mapping]] = ..., status: _Optional[_Union[ChannelStatus, str]] = ...) -> None: ...

class Channels(_message.Message):
    __slots__ = ("voice", "web_chat")
    VOICE_FIELD_NUMBER: _ClassVar[int]
    WEB_CHAT_FIELD_NUMBER: _ClassVar[int]
    voice: VoiceChannel
    web_chat: WebChatChannel
    def __init__(self, voice: _Optional[_Union[VoiceChannel, _Mapping]] = ..., web_chat: _Optional[_Union[WebChatChannel, _Mapping]] = ...) -> None: ...

class Channel_UpdateGreeting(_message.Message):
    __slots__ = ("channel_type", "greeting")
    CHANNEL_TYPE_FIELD_NUMBER: _ClassVar[int]
    GREETING_FIELD_NUMBER: _ClassVar[int]
    channel_type: ChannelType
    greeting: _agent_settings_pb2.Greeting_UpdateGreeting
    def __init__(self, channel_type: _Optional[_Union[ChannelType, str]] = ..., greeting: _Optional[_Union[_agent_settings_pb2.Greeting_UpdateGreeting, _Mapping]] = ...) -> None: ...

class Channel_UpdateStylePrompt(_message.Message):
    __slots__ = ("channel_type", "style_prompt")
    CHANNEL_TYPE_FIELD_NUMBER: _ClassVar[int]
    STYLE_PROMPT_FIELD_NUMBER: _ClassVar[int]
    channel_type: ChannelType
    style_prompt: StylePrompt_UpdateStylePrompt
    def __init__(self, channel_type: _Optional[_Union[ChannelType, str]] = ..., style_prompt: _Optional[_Union[StylePrompt_UpdateStylePrompt, _Mapping]] = ...) -> None: ...

class Channel_UpdateSafetyFilters(_message.Message):
    __slots__ = ("channel_type", "safety_filters")
    CHANNEL_TYPE_FIELD_NUMBER: _ClassVar[int]
    SAFETY_FILTERS_FIELD_NUMBER: _ClassVar[int]
    channel_type: ChannelType
    safety_filters: _content_filter_settings_pb2.ContentFilterSettings_UpdateContentFilterSettings
    def __init__(self, channel_type: _Optional[_Union[ChannelType, str]] = ..., safety_filters: _Optional[_Union[_content_filter_settings_pb2.ContentFilterSettings_UpdateContentFilterSettings, _Mapping]] = ...) -> None: ...

class Channel_UpdateLLMSettings(_message.Message):
    __slots__ = ("channel_type", "llm_settings")
    CHANNEL_TYPE_FIELD_NUMBER: _ClassVar[int]
    LLM_SETTINGS_FIELD_NUMBER: _ClassVar[int]
    channel_type: ChannelType
    llm_settings: _llm_settings_pb2.LLMSettings_UpdateLLMSettings
    def __init__(self, channel_type: _Optional[_Union[ChannelType, str]] = ..., llm_settings: _Optional[_Union[_llm_settings_pb2.LLMSettings_UpdateLLMSettings, _Mapping]] = ...) -> None: ...

class Channel_UpdateStatus(_message.Message):
    __slots__ = ("webchat",)
    WEBCHAT_FIELD_NUMBER: _ClassVar[int]
    webchat: WebChatChannel_UpdateStatus
    def __init__(self, webchat: _Optional[_Union[WebChatChannel_UpdateStatus, _Mapping]] = ...) -> None: ...

class VoiceChannel_UpdateASRSettings(_message.Message):
    __slots__ = ("asr_settings",)
    ASR_SETTINGS_FIELD_NUMBER: _ClassVar[int]
    asr_settings: _asr_settings_pb2.ASRSettings_UpdateASRSettings
    def __init__(self, asr_settings: _Optional[_Union[_asr_settings_pb2.ASRSettings_UpdateASRSettings, _Mapping]] = ...) -> None: ...

class VoiceChannel_UpdateDisclaimer(_message.Message):
    __slots__ = ("disclaimer",)
    DISCLAIMER_FIELD_NUMBER: _ClassVar[int]
    disclaimer: _agent_settings_pb2.DisclaimerMessage_UpdateDisclaimerMessage
    def __init__(self, disclaimer: _Optional[_Union[_agent_settings_pb2.DisclaimerMessage_UpdateDisclaimerMessage, _Mapping]] = ...) -> None: ...

class WebChatChannel_UpdateStatus(_message.Message):
    __slots__ = ("status",)
    STATUS_FIELD_NUMBER: _ClassVar[int]
    status: ChannelStatus
    def __init__(self, status: _Optional[_Union[ChannelStatus, str]] = ...) -> None: ...



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: functions.proto
# Protobuf Python Version: 4.25.5
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0f\x66unctions.proto\x12\tfunctions\x1a\x1fgoogle/protobuf/timestamp.proto\"\xd8\x01\n\x11\x46unctionParameter\x12\n\n\x02id\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x03 \x01(\t\x12\x0c\n\x04type\x18\x04 \x01(\t\x12.\n\ncreated_at\x18\x05 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x12\n\ncreated_by\x18\x06 \x01(\t\x12.\n\nupdated_at\x18\x07 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x12\n\nupdated_by\x18\x08 \x01(\t\">\n\rFunctionError\x12\x0e\n\x06lineno\x18\x01 \x01(\x05\x12\x0f\n\x07message\x18\x02 \x01(\t\x12\x0c\n\x04text\x18\x03 \x01(\t\"\xa1\x01\n\x1f\x46unctionDelayResponseReferences\x12L\n\tvariables\x18\x01 \x03(\x0b\x32\x39.functions.FunctionDelayResponseReferences.VariablesEntry\x1a\x30\n\x0eVariablesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x08:\x02\x38\x01\"\xa2\x02\n\x15\x46unctionDelayResponse\x12\n\n\x02id\x18\x01 \x01(\t\x12\x0f\n\x07message\x18\x02 \x01(\t\x12\x10\n\x08\x64uration\x18\x03 \x01(\x05\x12.\n\ncreated_at\x18\x04 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x12\n\ncreated_by\x18\x05 \x01(\t\x12.\n\nupdated_at\x18\x06 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x12\n\nupdated_by\x18\x07 \x01(\t\x12\x43\n\nreferences\x18\x08 \x01(\x0b\x32*.functions.FunctionDelayResponseReferencesH\x00\x88\x01\x01\x42\r\n\x0b_references\"\x95\x02\n\x16\x46unctionLatencyControl\x12\x0f\n\x07\x65nabled\x18\x01 \x01(\x08\x12\x39\n\x0f\x64\x65lay_responses\x18\x02 \x03(\x0b\x32 .functions.FunctionDelayResponse\x12\x15\n\rinitial_delay\x18\x03 \x01(\x05\x12\x10\n\x08interval\x18\x04 \x01(\x05\x12.\n\ncreated_at\x18\x05 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x12\n\ncreated_by\x18\x06 \x01(\t\x12.\n\nupdated_at\x18\x07 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x12\n\nupdated_by\x18\x08 \x01(\t\"\xd5\x04\n\x12\x46unctionReferences\x12@\n\nflow_steps\x18\x01 \x03(\x0b\x32,.functions.FunctionReferences.FlowStepsEntry\x12\x39\n\x06topics\x18\x02 \x03(\x0b\x32).functions.FunctionReferences.TopicsEntry\x12\x46\n\rstop_keywords\x18\x03 \x03(\x0b\x32/.functions.FunctionReferences.StopKeywordsEntry\x12?\n\tbehaviour\x18\x04 \x03(\x0b\x32,.functions.FunctionReferences.BehaviourEntry\x12?\n\tvariables\x18\x05 \x03(\x0b\x32,.functions.FunctionReferences.VariablesEntry\x1a\x30\n\x0e\x46lowStepsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x08:\x02\x38\x01\x1a-\n\x0bTopicsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x08:\x02\x38\x01\x1a\x33\n\x11StopKeywordsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x08:\x02\x38\x01\x1a\x30\n\x0e\x42\x65haviourEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x08:\x02\x38\x01\x1a\x30\n\x0eVariablesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x08:\x02\x38\x01\"\xac\x03\n\x08\x46unction\x12\n\n\x02id\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x03 \x01(\t\x12\x30\n\nparameters\x18\x04 \x03(\x0b\x32\x1c.functions.FunctionParameter\x12\x0c\n\x04\x63ode\x18\x05 \x01(\t\x12(\n\x06\x65rrors\x18\x06 \x03(\x0b\x32\x18.functions.FunctionError\x12:\n\x0flatency_control\x18\x07 \x01(\x0b\x32!.functions.FunctionLatencyControl\x12.\n\ncreated_at\x18\x08 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x12\n\ncreated_by\x18\t \x01(\t\x12.\n\nupdated_at\x18\n \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x12\n\nupdated_by\x18\x0b \x01(\t\x12\x10\n\x08\x61rchived\x18\x0c \x01(\x08\x12\x31\n\nreferences\x18\r \x01(\x0b\x32\x1d.functions.FunctionReferences\"3\n\tFunctions\x12&\n\tfunctions\x18\x01 \x03(\x0b\x32\x13.functions.Function\"V\n\x17\x46unctionParameterUpdate\x12\n\n\x02id\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x03 \x01(\t\x12\x0c\n\x04type\x18\x04 \x01(\t\"\x93\x01\n\x1c\x46unctionCreateLatencyControl\x12\x0f\n\x07\x65nabled\x18\x01 \x01(\x08\x12\x39\n\x0f\x64\x65lay_responses\x18\x02 \x03(\x0b\x32 .functions.FunctionDelayResponse\x12\x15\n\rinitial_delay\x18\x03 \x01(\x05\x12\x10\n\x08interval\x18\x04 \x01(\x05\"\xfe\x02\n\x17\x46unction_CreateFunction\x12\n\n\x02id\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x03 \x01(\t\x12\x36\n\nparameters\x18\x04 \x03(\x0b\x32\".functions.FunctionParameterUpdate\x12\x0c\n\x04\x63ode\x18\x05 \x01(\t\x12(\n\x06\x65rrors\x18\x06 \x03(\x0b\x32\x18.functions.FunctionError\x12\x45\n\x0flatency_control\x18\x07 \x01(\x0b\x32\'.functions.FunctionCreateLatencyControlH\x00\x88\x01\x01\x12\x36\n\nreferences\x18\x08 \x01(\x0b\x32\x1d.functions.FunctionReferencesH\x01\x88\x01\x01\x12\x15\n\x08\x61rchived\x18\t \x01(\x08H\x02\x88\x01\x01\x42\x12\n\x10_latency_controlB\r\n\x0b_referencesB\x0b\n\t_archived\"J\n\x10ParametersUpdate\x12\x36\n\nparameters\x18\x01 \x03(\x0b\x32\".functions.FunctionParameterUpdate\"8\n\x0c\x45rrorsUpdate\x12(\n\x06\x65rrors\x18\x01 \x03(\x0b\x32\x18.functions.FunctionError\"\xf0\x02\n\x17\x46unction_UpdateFunction\x12\n\n\x02id\x18\x01 \x01(\t\x12\x11\n\x04name\x18\x02 \x01(\tH\x00\x88\x01\x01\x12\x18\n\x0b\x64\x65scription\x18\x03 \x01(\tH\x01\x88\x01\x01\x12\x34\n\nparameters\x18\x04 \x01(\x0b\x32\x1b.functions.ParametersUpdateH\x02\x88\x01\x01\x12\x11\n\x04\x63ode\x18\x05 \x01(\tH\x03\x88\x01\x01\x12,\n\x06\x65rrors\x18\x06 \x01(\x0b\x32\x17.functions.ErrorsUpdateH\x04\x88\x01\x01\x12\x36\n\nreferences\x18\x07 \x01(\x0b\x32\x1d.functions.FunctionReferencesH\x05\x88\x01\x01\x12\x15\n\x08\x61rchived\x18\x08 \x01(\x08H\x06\x88\x01\x01\x42\x07\n\x05_nameB\x0e\n\x0c_descriptionB\r\n\x0b_parametersB\x07\n\x05_codeB\t\n\x07_errorsB\r\n\x0b_referencesB\x0b\n\t_archived\"%\n\x17\x46unction_DeleteFunction\x12\n\n\x02id\x18\x01 \x01(\t\"\x98\x01\n\x13\x44\x65layResponseUpdate\x12\n\n\x02id\x18\x01 \x01(\t\x12\x0f\n\x07message\x18\x02 \x01(\t\x12\x10\n\x08\x64uration\x18\x03 \x01(\x05\x12\x43\n\nreferences\x18\x04 \x01(\x0b\x32*.functions.FunctionDelayResponseReferencesH\x00\x88\x01\x01\x42\r\n\x0b_references\"O\n\x14\x44\x65layResponsesUpdate\x12\x37\n\x0f\x64\x65lay_responses\x18\x01 \x03(\x0b\x32\x1e.functions.DelayResponseUpdate\"\xea\x01\n\x1d\x46unction_UpdateLatencyControl\x12\x13\n\x0b\x66unction_id\x18\x01 \x01(\t\x12\x0f\n\x07\x65nabled\x18\x02 \x01(\x08\x12=\n\x0f\x64\x65lay_responses\x18\x03 \x01(\x0b\x32\x1f.functions.DelayResponsesUpdateH\x00\x88\x01\x01\x12\x1a\n\rinitial_delay\x18\x04 \x01(\x05H\x01\x88\x01\x01\x12\x15\n\x08interval\x18\x05 \x01(\x05H\x02\x88\x01\x01\x42\x12\n\x10_delay_responsesB\x10\n\x0e_initial_delayB\x0b\n\t_intervalb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'functions_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  _globals['_FUNCTIONDELAYRESPONSEREFERENCES_VARIABLESENTRY']._options = None
  _globals['_FUNCTIONDELAYRESPONSEREFERENCES_VARIABLESENTRY']._serialized_options = b'8\001'
  _globals['_FUNCTIONREFERENCES_FLOWSTEPSENTRY']._options = None
  _globals['_FUNCTIONREFERENCES_FLOWSTEPSENTRY']._serialized_options = b'8\001'
  _globals['_FUNCTIONREFERENCES_TOPICSENTRY']._options = None
  _globals['_FUNCTIONREFERENCES_TOPICSENTRY']._serialized_options = b'8\001'
  _globals['_FUNCTIONREFERENCES_STOPKEYWORDSENTRY']._options = None
  _globals['_FUNCTIONREFERENCES_STOPKEYWORDSENTRY']._serialized_options = b'8\001'
  _globals['_FUNCTIONREFERENCES_BEHAVIOURENTRY']._options = None
  _globals['_FUNCTIONREFERENCES_BEHAVIOURENTRY']._serialized_options = b'8\001'
  _globals['_FUNCTIONREFERENCES_VARIABLESENTRY']._options = None
  _globals['_FUNCTIONREFERENCES_VARIABLESENTRY']._serialized_options = b'8\001'
  _globals['_FUNCTIONPARAMETER']._serialized_start=64
  _globals['_FUNCTIONPARAMETER']._serialized_end=280
  _globals['_FUNCTIONERROR']._serialized_start=282
  _globals['_FUNCTIONERROR']._serialized_end=344
  _globals['_FUNCTIONDELAYRESPONSEREFERENCES']._serialized_start=347
  _globals['_FUNCTIONDELAYRESPONSEREFERENCES']._serialized_end=508
  _globals['_FUNCTIONDELAYRESPONSEREFERENCES_VARIABLESENTRY']._serialized_start=460
  _globals['_FUNCTIONDELAYRESPONSEREFERENCES_VARIABLESENTRY']._serialized_end=508
  _globals['_FUNCTIONDELAYRESPONSE']._serialized_start=511
  _globals['_FUNCTIONDELAYRESPONSE']._serialized_end=801
  _globals['_FUNCTIONLATENCYCONTROL']._serialized_start=804
  _globals['_FUNCTIONLATENCYCONTROL']._serialized_end=1081
  _globals['_FUNCTIONREFERENCES']._serialized_start=1084
  _globals['_FUNCTIONREFERENCES']._serialized_end=1681
  _globals['_FUNCTIONREFERENCES_FLOWSTEPSENTRY']._serialized_start=1433
  _globals['_FUNCTIONREFERENCES_FLOWSTEPSENTRY']._serialized_end=1481
  _globals['_FUNCTIONREFERENCES_TOPICSENTRY']._serialized_start=1483
  _globals['_FUNCTIONREFERENCES_TOPICSENTRY']._serialized_end=1528
  _globals['_FUNCTIONREFERENCES_STOPKEYWORDSENTRY']._serialized_start=1530
  _globals['_FUNCTIONREFERENCES_STOPKEYWORDSENTRY']._serialized_end=1581
  _globals['_FUNCTIONREFERENCES_BEHAVIOURENTRY']._serialized_start=1583
  _globals['_FUNCTIONREFERENCES_BEHAVIOURENTRY']._serialized_end=1631
  _globals['_FUNCTIONREFERENCES_VARIABLESENTRY']._serialized_start=460
  _globals['_FUNCTIONREFERENCES_VARIABLESENTRY']._serialized_end=508
  _globals['_FUNCTION']._serialized_start=1684
  _globals['_FUNCTION']._serialized_end=2112
  _globals['_FUNCTIONS']._serialized_start=2114
  _globals['_FUNCTIONS']._serialized_end=2165
  _globals['_FUNCTIONPARAMETERUPDATE']._serialized_start=2167
  _globals['_FUNCTIONPARAMETERUPDATE']._serialized_end=2253
  _globals['_FUNCTIONCREATELATENCYCONTROL']._serialized_start=2256
  _globals['_FUNCTIONCREATELATENCYCONTROL']._serialized_end=2403
  _globals['_FUNCTION_CREATEFUNCTION']._serialized_start=2406
  _globals['_FUNCTION_CREATEFUNCTION']._serialized_end=2788
  _globals['_PARAMETERSUPDATE']._serialized_start=2790
  _globals['_PARAMETERSUPDATE']._serialized_end=2864
  _globals['_ERRORSUPDATE']._serialized_start=2866
  _globals['_ERRORSUPDATE']._serialized_end=2922
  _globals['_FUNCTION_UPDATEFUNCTION']._serialized_start=2925
  _globals['_FUNCTION_UPDATEFUNCTION']._serialized_end=3293
  _globals['_FUNCTION_DELETEFUNCTION']._serialized_start=3295
  _globals['_FUNCTION_DELETEFUNCTION']._serialized_end=3332
  _globals['_DELAYRESPONSEUPDATE']._serialized_start=3335
  _globals['_DELAYRESPONSEUPDATE']._serialized_end=3487
  _globals['_DELAYRESPONSESUPDATE']._serialized_start=3489
  _globals['_DELAYRESPONSESUPDATE']._serialized_end=3568
  _globals['_FUNCTION_UPDATELATENCYCONTROL']._serialized_start=3571
  _globals['_FUNCTION_UPDATELATENCYCONTROL']._serialized_end=3805
# @@protoc_insertion_point(module_scope)



"""Handling and managing Agent Studio Keyphrase Boosting (ASR Keyphrases)

Copyright PolyAI Limited
"""

import os
from dataclasses import dataclass
from typing import ClassVar, Optional

import poly.resources.resource_utils as utils
from poly.handlers.protobuf.keyphrase_boosting_pb2 import (
    KeyphraseBoosting_CreateKeyphrase,
    KeyphraseBoosting_DeleteKeyphrase,
    KeyphraseBoosting_UpdateKeyphrase,
)
from poly.resources.resource import MultiResourceYamlResource

VALID_LEVELS = ("default", "boosted", "maximum")


@dataclass
class KeyphraseBoosting(MultiResourceYamlResource):
    """Dataclass representing an ASR Keyphrase Boosting entry"""

    keyphrase: str
    level: str
    top_level_name: ClassVar[str] = "keyphrases"

    def __init__(
        self,
        *,
        resource_id: Optional[str] = None,
        name: str = "",
        keyphrase: str = "",
        level: str = "default",
    ):
        self.resource_id = resource_id
        self.name = name
        self.keyphrase = keyphrase
        self.level = level.lower()

    @property
    def file_path(self) -> str:
        path_safe_name = utils.clean_name(self.name, lowercase=False)
        return os.path.join(
            "voice",
            "speech_recognition",
            "keyphrase_boosting.yaml",
            self.top_level_name,
            path_safe_name,
        )

    def to_yaml_dict(self) -> dict:
        return {
            "keyphrase": self.keyphrase,
            "level": self.level,
        }

    @classmethod
    def from_yaml_dict(
        cls, yaml_dict: dict, resource_id: str, name: str, **kwargs
    ) -> "KeyphraseBoosting":
        return cls(
            resource_id=resource_id,
            name=yaml_dict.get("keyphrase", name),
            keyphrase=yaml_dict.get("keyphrase", ""),
            level=yaml_dict.get("level", "default").lower(),
        )

    @property
    def command_type(self) -> str:
        return "keyphrase_boosting"

    def build_create_proto(self) -> KeyphraseBoosting_CreateKeyphrase:
        return KeyphraseBoosting_CreateKeyphrase(
            id=self.resource_id,
            keyphrase=self.keyphrase,
            level=self.level,
        )

    def build_update_proto(self) -> KeyphraseBoosting_UpdateKeyphrase:
        return KeyphraseBoosting_UpdateKeyphrase(
            id=self.resource_id,
            keyphrase=self.keyphrase,
            level=self.level,
        )

    def build_delete_proto(self) -> KeyphraseBoosting_DeleteKeyphrase:
        return KeyphraseBoosting_DeleteKeyphrase(
            id=self.resource_id,
        )

    def validate(self, **kwargs) -> None:
        if not self.keyphrase:
            raise ValueError("Keyphrase is required")
        if self.level not in VALID_LEVELS:
            raise ValueError(
                f"Invalid level '{self.level}'. Must be one of: {', '.join(VALID_LEVELS)}"
            )

    @staticmethod
    def _find_matching(yaml_list, resource_clean_name) -> Optional[dict]:
        for r in yaml_list:
            if utils.clean_name(r.get("keyphrase") or "", lowercase=False) == resource_clean_name:
                return r
        return None

    @staticmethod
    def discover_resources(base_path: str) -> list[str]:
        # Must match file_path: voice/speech_recognition/keyphrase_boosting.yaml
        # Also check legacy path: speech_recognition/keyphrase_boosting.yaml
        for rel_path in (
            os.path.join("voice", "speech_recognition", "keyphrase_boosting.yaml"),
            os.path.join("speech_recognition", "keyphrase_boosting.yaml"),
        ):
            yaml_path = os.path.join(base_path, rel_path)
            if os.path.exists(yaml_path):
                break
        else:
            return []

        discovered: list[str] = []
        yaml_dict = KeyphraseBoosting._get_top_level_data(yaml_path)
        keyphrases: list[dict] = yaml_dict.get("keyphrases", []) if yaml_dict else []

        for kp in keyphrases:
            name = kp.get("keyphrase")
            if not name:
                continue
            path_safe_name = utils.clean_name(name, lowercase=False)
            discovered.append(
                os.path.join(yaml_path, KeyphraseBoosting.top_level_name, path_safe_name)
            )

        return discovered



"""Handling and managing an Agent Studio Flows

Copyright PolyAI Limited
"""

import os
import re
import uuid
from abc import ABC
from dataclasses import dataclass, field
from enum import Enum
from functools import cached_property
from typing import Optional

import poly.resources.resource_utils as utils
from poly.handlers.protobuf.flows_pb2 import (
    AdvancedStepCondition,
    ConditionDetails,
    CreateAdvancedStep,
    CreateFunctionStep,
    CreateFunctionStepDefinition,
    CreateNoCodeCondition,
    CreateNoCodeStep,
    CreateStep,
    DeleteNoCodeCondition,
    DeleteNoCodeStep,
    DeleteStep,
    ExitFlowCondition,
    Flow_CreateFlow,
    Flow_CreateStep,
    Flow_DeleteFlow,
    Flow_DeleteStep,
    Flow_UpdateFlow,
    Flow_UpdateStep,
    Flow_UpdateStepAsrConfig,
    Flow_UpdateStepDtmfConfig,
    FunctionStepCondition,
    NoCodeStepCondition,
    StepAsrConfig,
    StepAsrConfigUpdate,
    StepDtmfConfig,
    StepDtmfConfigUpdate,
    StepPosition,
    UpdateAdvancedStep,
    UpdateAsrKeywords,
    UpdateFunctionStep,
    UpdateFunctionStepDefinition,
    UpdateNoCodeCondition,
    UpdateNoCodeStep,
    UpdateStep,
)
from poly.resources.entities import Entity
from poly.resources.function import Function, FunctionType
from poly.resources.resource import (
    ResourceMapping,
    SubResource,
    YamlResource,
)

FUNCTION_REGEX = re.compile(r"{{f[nt]:([\w-]+)}}")
# Flow step names: alphanumeric, extended Latin (C0–024F, 1E00–1EFF), and _ &,/.-
FLOW_STEP_NAME_PATTERN = re.compile(r"^[a-zA-Z0-9\u00C0-\u024F\u1E00-\u1EFF_ &,/.\-]+$")
FLOW_REFERENCES = [
    "global_functions",
    "sms",
    "handoff",
    "attributes",
    "transition_functions",
    "entities",
    "variables",
]
NO_CODE_STEP_REFERENCES = [
    "attributes",
    "entities",
    "variables",
]


@dataclass
class FlowConfig(YamlResource):
    """Flow configuration resource."""

    description: str = field(default="")
    start_step: str = field(default="")

    # For creating:
    functions: list[Function] = field(default_factory=list, repr=False, init=False)
    steps: list["FlowStep"] = field(default_factory=list, repr=False, init=False)

    @cached_property
    def file_path(self) -> str:
        """File path for the resource."""
        return os.path.join("flows", utils.clean_name(self.name), "flow_config.yaml")

    def to_yaml_dict(self) -> dict:
        """Return a dictionary suitable for YAML serialization."""
        return {
            "name": self.name,
            "description": self.description,
            "start_step": self.start_step,
        }

    @classmethod
    def from_yaml_dict(cls, yaml_data: dict, resource_id: str, name: str, **kwargs) -> "FlowConfig":
        """Create an instance from YAML data and identity fields."""
        return cls(
            resource_id=resource_id,
            name=yaml_data.get("name", ""),
            description=(yaml_data.get("description") or "").strip(),
            start_step=yaml_data.get("start_step", ""),
        )

    @classmethod
    def read_local_resource(
        cls, file_path: str, resource_id: str, resource_name: str, **kwargs
    ) -> "FlowConfig":
        """Read a local YAML resource from the given file path."""
        flow_config: FlowConfig = super().read_local_resource(
            file_path, resource_id=resource_id, resource_name=resource_name, **kwargs
        )
        # Just check flow folder name (one level up from file) matches flow name
        file_path_part_flow_folder = utils.get_flow_name_from_path(file_path)
        expected_flow_folder = utils.clean_name(flow_config.name)

        if file_path_part_flow_folder != expected_flow_folder:
            raise ValueError(
                f"Flow folder name does not match flow name in config. {file_path_part_flow_folder} != {expected_flow_folder}"
            )
        return flow_config

    @staticmethod
    def to_pretty_dict(
        d: dict,
        resource_name: str = None,
        resource_mappings: list[ResourceMapping] = None,
        **kwargs,
    ) -> dict:
        """Return dict with start_step ID replaced by name."""
        d = d.copy()
        start_step_id = d.get("start_step")
        if start_step_id:
            for resource in resource_mappings or []:
                if (
                    issubclass(resource.resource_type, BaseFlowStep)
                    and resource_name == resource.flow_name
                    and resource.resource_id.removeprefix(resource.flow_name + "_") == start_step_id
                ):
                    d["start_step"] = resource.resource_name
                    break
        return d

    @staticmethod
    def from_pretty(
        contents: str,
        resource_name: str = None,
        resource_mappings: list[ResourceMapping] = None,
        **kwargs,
    ) -> str:
        """Replace resource names with resource IDs in the provided contents."""
        # Replace start_step name with ID if step from same flow
        try:
            yaml_dict = utils.load_yaml(contents) or {}
        except Exception as e:
            raise ValueError(f"Error loading YAML content for {resource_name}") from e

        start_step_name = yaml_dict.get("start_step")
        if start_step_name:
            for resource in resource_mappings or []:
                if (
                    issubclass(resource.resource_type, BaseFlowStep)
                    and resource.flow_name == resource_name
                    and resource.resource_name == start_step_name
                ):
                    yaml_dict["start_step"] = resource.resource_id.removeprefix(
                        resource.flow_name + "_"
                    )
                    break
        return utils.dump_yaml(yaml_dict)

    def validate(self, resource_mappings: list[ResourceMapping] = None, **kwargs):
        """Validate the flow config resource."""
        if not self.start_step:
            raise ValueError("Start step cannot be empty.")

        # Check flow step exists in resource mappings
        found_step = False
        expected_step_resource_id = f"{self.name}_{self.start_step}"
        for resource in resource_mappings or []:
            if (
                issubclass(resource.resource_type, BaseFlowStep)
                and resource.flow_name == self.name
                and resource.resource_id == expected_step_resource_id
            ):
                found_step = True
                break

        if not found_step:
            raise ValueError(f"Start step '{self.start_step}' not found.")

        # Check description exists
        if not self.description:
            raise ValueError("Description cannot be empty.")

        if self.description != self.description.strip():
            raise ValueError("Description cannot contain leading or trailing whitespace.")

    def build_update_proto(
        self,
    ) -> Flow_UpdateFlow:
        """Create a proto for updating the resource."""
        return Flow_UpdateFlow(
            flow_id=self.resource_id,
            name=self.name,
            description=self.description,
            start_step_id=self.start_step,
        )

    def build_delete_proto(self):
        """Create a proto for deleting the resource."""
        return Flow_DeleteFlow(flow_id=self.resource_id)

    def build_create_proto(self):
        """Create a proto for creating the resource."""
        functions = self.functions or []
        all_steps = self.steps or []

        transition_functions = [
            function.build_create_proto().transition_function for function in functions
        ]

        steps = [step.build_create_proto() for step in all_steps]

        return Flow_CreateFlow(
            id=self.resource_id,
            name=self.name,
            description=self.description,
            start_step_id=self.start_step,
            transition_functions=transition_functions,
            no_code_steps=[step for step in steps if isinstance(step, CreateNoCodeStep)],
            steps=[step.step for step in steps if isinstance(step, Flow_CreateStep)],
        )

    @property
    def command_type(self) -> str:
        """Get the update type for updating the resource."""
        return "flow"

    @staticmethod
    def discover_resources(base_path: str) -> list[str]:
        """Discover resources of this type in the given base path.

        Args:
            base_path (str): The base path to search for resources.

        Returns:
            list[str]: A list of file paths of discovered resources.
        """
        flows_path = os.path.join(base_path, "flows")
        discovered_flow_configs = []
        if not os.path.exists(flows_path):
            return discovered_flow_configs

        # Find all the flow configs that match the pattern flows/<flow_name_formatted>/flow_config.yaml
        for flow_name_formatted in os.listdir(flows_path):
            flow_folder_path = os.path.join(flows_path, flow_name_formatted)

            # Skip if not a directory (e.g., files in flows/)
            if not os.path.isdir(flow_folder_path):
                continue

            for flow_config_path in os.listdir(flow_folder_path):
                if flow_config_path == "flow_config.yaml":
                    discovered_flow_configs.append(
                        os.path.join(flows_path, flow_name_formatted, flow_config_path)
                    )

        return discovered_flow_configs


class StepType(str, Enum):
    """Enum for step types."""

    ADVANCED_STEP = "advanced_step"
    DEFAULT_STEP = "default_step"
    FUNCTION_STEP = "function_step"


@dataclass
class BaseFlowStep(ABC):
    # Store step_id as not unique across flows
    step_id: str
    flow_id: str
    flow_name: str
    step_type: StepType
    position: dict[str, float]


@dataclass
class FlowStep(BaseFlowStep, YamlResource):
    """Flow step resource."""

    asr_biasing: Optional["ASRBiasing"]
    dtmf_config: Optional["DTMFConfig"]
    conditions: Optional[list["Condition"]]
    extracted_entities: Optional[list[str]]
    prompt: str

    def __init__(
        self,
        resource_id: str,
        name: str,
        step_id: str,
        flow_id: str,
        flow_name: str,
        step_type: "str | StepType",
        prompt: str,
        asr_biasing: Optional["ASRBiasing | dict"] = None,
        dtmf_config: Optional["DTMFConfig | dict"] = None,
        conditions: Optional[list["Condition | dict"]] = None,
        extracted_entities: Optional[list[str]] = None,
        position: Optional[dict[str, float]] = None,
    ):
        self.resource_id = resource_id
        self.name = name
        self.step_id = step_id
        self.flow_id = flow_id
        self.flow_name = flow_name
        if not step_type:
            raise ValueError("step_type is required for FlowStep")
        self.step_type = StepType(step_type) if isinstance(step_type, str) else step_type

        if isinstance(asr_biasing, ASRBiasing):
            self.asr_biasing = asr_biasing
        elif asr_biasing is not None:
            # resource_id and name are set internally by ASRBiasing.__init__;
            asr_biasing = {k: v for k, v in asr_biasing.items() if k not in ("resource_id", "name")}
            asr_biasing["step_id"] = self.step_id
            asr_biasing["flow_id"] = self.flow_id
            self.asr_biasing = ASRBiasing(**asr_biasing)
        else:
            self.asr_biasing = None

        if isinstance(dtmf_config, DTMFConfig):
            self.dtmf_config = dtmf_config
        elif dtmf_config is not None:
            # resource_id and name are set internally by DTMFConfig.__init__;
            dtmf_config = {k: v for k, v in dtmf_config.items() if k not in ("resource_id", "name")}
            dtmf_config["step_id"] = self.step_id
            dtmf_config["flow_id"] = self.flow_id
            self.dtmf_config = DTMFConfig(**dtmf_config)
        else:
            self.dtmf_config = None

        self.extracted_entities = extracted_entities or []
        self.conditions = [
            Condition(**condition) if not isinstance(condition, Condition) else condition
            for condition in (conditions or [])
        ]
        self.prompt = prompt
        self.position = position or {}

    def to_yaml_dict(self) -> dict:
        """Return a dictionary suitable for YAML serialization."""
        output = {
            "step_type": self.step_type.value,
            "name": self.name,
        }
        if self.step_type == StepType.ADVANCED_STEP:
            output["asr_biasing"] = self.asr_biasing.to_yaml_dict() if self.asr_biasing else {}
            output["dtmf_config"] = self.dtmf_config.to_yaml_dict() if self.dtmf_config else {}

        if self.step_type == StepType.DEFAULT_STEP:
            output["conditions"] = [condition.to_yaml_dict() for condition in self.conditions]
            output["extracted_entities"] = self.extracted_entities

        output["prompt"] = self.prompt
        return output

    @classmethod
    def from_yaml_dict(
        cls,
        yaml_dict: dict,
        resource_id: str,
        file_name: str,
        flow_id: str,
        flow_name: str,
        name: str = "",
        known_position: dict[str, float] = None,
        known_conditions: list["Condition"] = None,
        resource_mappings: list[ResourceMapping] = None,
        **kwargs,
    ) -> "YamlResource":
        """Create an instance from YAML data and identity fields."""
        # Map conditions by name
        known_conditions = known_conditions or []
        condition_name_map = {cond.name: cond for cond in known_conditions}

        step_id = resource_id.removeprefix(f"{flow_name}_")

        conditions = []
        for condition_yaml in yaml_dict.get("conditions", []):
            condition_name = condition_yaml.get("name")
            known_condition = condition_name_map.get(condition_name)

            # Find Child Step to infer condition type if needed
            child_step_type = None
            if child_step_id := condition_yaml.get("child_step"):
                for resource in resource_mappings or []:
                    if (
                        issubclass(resource.resource_type, BaseFlowStep)
                        and resource.flow_name == flow_name
                        and resource.resource_id.removeprefix(f"{flow_name}_") == child_step_id
                    ):
                        if issubclass(resource.resource_type, FunctionStep):
                            child_step_type = StepType.FUNCTION_STEP
                            break

                        child_step_contents = cls.read_to_raw(
                            resource.file_path,
                            resource_mappings=resource_mappings,
                            flow_name=flow_name,
                        )
                        child_step_yaml = utils.load_yaml(child_step_contents)
                        child_step_type = StepType(child_step_yaml.get("step_type"))
                        break

            conditions.append(
                Condition.from_yaml_dict(
                    condition_yaml,
                    flow_id=flow_id,
                    step_id=step_id,
                    resource_id=(
                        known_condition.resource_id
                        if known_condition
                        else f"CONDITION-{uuid.uuid4().hex[:8]}"
                    ),
                    position=known_condition.position if known_condition else None,
                    ingress=known_condition.ingress if known_condition else None,
                    exit_flow_position=known_condition.exit_flow_position
                    if known_condition
                    else None,
                    child_step_type=child_step_type,
                )
            )
            if known_condition:
                del condition_name_map[condition_name]

        yaml_name = yaml_dict.get("name")

        if file_name != utils.clean_name(yaml_name):
            raise ValueError(
                f"Step name {yaml_name} in file {file_name}.yaml does not match clean version of name expected: {utils.clean_name(yaml_name)}.yaml"
            )

        step_type = StepType(yaml_dict.get("step_type"))
        asr_biasing = yaml_dict.get("asr_biasing", {})
        dtmf_config = yaml_dict.get("dtmf_config", {})
        extracted_entities = yaml_dict.get("extracted_entities", [])
        if step_type == StepType.DEFAULT_STEP:
            # ASR and DTMF not applicable
            asr_biasing = {}
            dtmf_config = {}
        elif step_type == StepType.ADVANCED_STEP:
            # Conditions not applicable
            conditions = []

        return cls(
            resource_id=resource_id,
            step_id=step_id,
            name=yaml_dict.get("name"),
            flow_id=flow_id,
            flow_name=flow_name,
            step_type=step_type,
            asr_biasing=asr_biasing,
            dtmf_config=dtmf_config,
            prompt=yaml_dict.get("prompt"),
            conditions=conditions,
            position=known_position,
            extracted_entities=extracted_entities,
        )

    @staticmethod
    def to_pretty_dict(
        d: dict,
        file_path: str = None,
        resource_mappings: list[ResourceMapping] = None,
        **kwargs,
    ) -> dict:
        """Return dict with resource IDs replaced by names."""
        d = d.copy()
        flow_folder_name = utils.get_flow_name_from_path(file_path)

        if not flow_folder_name:
            raise ValueError(
                f"Flow folder name could not be determined from file_path: {file_path}"
            )

        if prompt := d.get("prompt"):
            d["prompt"] = utils.replace_resource_ids_with_names(
                prompt, resource_mappings or [], flow_folder_name=flow_folder_name
            )

        entity_mappings = {
            resource.resource_id: resource.resource_name
            for resource in resource_mappings or []
            if resource.resource_type == Entity
        }

        if extracted_entities := d.get("extracted_entities"):
            d["extracted_entities"] = [
                entity_mappings.get(entity_id, entity_id) for entity_id in extracted_entities
            ]

        if conditions := d.get("conditions"):
            for condition in conditions:
                if child_step_id := condition.get("child_step"):
                    for resource in resource_mappings or []:
                        if (
                            issubclass(resource.resource_type, BaseFlowStep)
                            and flow_folder_name in resource.file_path.split(os.sep)
                            and resource.resource_id.removeprefix(resource.flow_name + "_")
                            == child_step_id
                        ):
                            condition["child_step"] = resource.resource_name
                            break

                if required_entities := condition.get("required_entities"):
                    condition["required_entities"] = [
                        entity_mappings.get(entity_id, entity_id) for entity_id in required_entities
                    ]
        return d

    @staticmethod
    def from_pretty(
        contents: str,
        file_path: str = None,
        resource_mappings: list[ResourceMapping] = None,
        **kwargs,
    ) -> str:
        """Replace resource names with resource IDs in the provided contents."""
        flow_folder_name = utils.get_flow_name_from_path(file_path)

        if not flow_folder_name:
            raise ValueError("flow_name could not be determined from file_path")

        try:
            yaml_dict = utils.load_yaml(contents) or {}
        except Exception as e:
            raise ValueError(f"Error loading YAML content for {file_path}") from e

        if prompt := yaml_dict.get("prompt"):
            yaml_dict["prompt"] = utils.replace_resource_names_with_ids(
                prompt, resource_mappings or [], flow_folder_name=flow_folder_name
            )

        entity_mappings = {
            resource.resource_name: resource.resource_id
            for resource in resource_mappings or []
            if resource.resource_type == Entity
        }

        if extracted_entities := yaml_dict.get("extracted_entities"):
            new_requested_entities = [
                entity_mappings.get(entity_name, entity_name) for entity_name in extracted_entities
            ]
            yaml_dict["extracted_entities"] = new_requested_entities

        # Replace child name with ID if step from same flow
        if conditions := yaml_dict.get("conditions"):
            for condition in conditions:
                if child_step_name := condition.get("child_step"):
                    for resource in resource_mappings or []:
                        if (
                            issubclass(resource.resource_type, BaseFlowStep)
                            and flow_folder_name in resource.file_path.split(os.sep)
                            and resource.resource_name == child_step_name
                        ):
                            condition["child_step"] = resource.resource_id.removeprefix(
                                resource.flow_name + "_"
                            )
                            break

                if required_entities := condition.get("required_entities"):
                    new_required_entities = [
                        entity_mappings.get(entity_name, entity_name)
                        for entity_name in required_entities
                    ]
                    condition["required_entities"] = new_required_entities

        return utils.dump_yaml(yaml_dict)

    @cached_property
    def file_path(self) -> str:
        """File path for the resource."""
        return os.path.join(
            "flows",
            utils.clean_name(self.flow_name),
            "steps",
            f"{utils.clean_name(self.name)}.yaml",
        )

    @classmethod
    def read_local_resource(
        cls,
        file_path: str,
        resource_id: str,
        resource_name: str,
        resource_mappings: list[ResourceMapping],
        known_conditions: list["Condition"] = None,
        known_position: dict[str, float] = None,
        **kwargs,
    ) -> "YamlResource":
        """Read a local YAML resource from the given file path."""
        flow_folder_name = utils.get_flow_name_from_path(file_path)

        # Extract flow_id from resource mappings
        flow_id, flow_name = utils.get_flow_id_from_flow_name(flow_folder_name, resource_mappings)

        content = cls.read_to_raw(
            file_path,
            resource_mappings=resource_mappings,
            flow_name=flow_name,
            **kwargs,
        )
        try:
            yaml_dict = utils.load_yaml(content) or {}
        except Exception as e:
            raise ValueError(f"Error loading YAML file: {file_path}") from e

        # Get file name from file_path
        file_name = os.path.splitext(os.path.basename(file_path))[0]

        return cls.from_yaml_dict(
            yaml_dict,
            resource_id=resource_id,
            file_name=file_name,
            flow_id=flow_id,
            flow_name=flow_name,
            known_conditions=known_conditions,
            known_position=known_position,
            resource_mappings=resource_mappings,
        )

    def validate(self, resource_mappings: list[ResourceMapping] = None, **kwargs):
        """Validate the flow step resource."""
        if not self.name:
            raise ValueError("Name cannot be empty.")

        if not FLOW_STEP_NAME_PATTERN.fullmatch(self.name):
            raise ValueError(
                "Name must contain only letters (including accented), numbers, and _ & , / . -"
            )

        if self.prompt is None or not self.prompt.strip():
            raise ValueError("Prompt cannot be empty.")

        # Check flow config exists in resource mappings
        found_flow = False

        if not self.flow_id:
            raise ValueError("Flow ID cannot be empty.")

        for resource in resource_mappings or []:
            if resource.resource_type == FlowConfig and resource.resource_id == self.flow_id:
                found_flow = True
                break

        if not found_flow:
            raise ValueError("Flow config not found.")

        if self.step_type not in StepType:
            raise ValueError(
                f"Invalid step type: {self.step_type}. Valid types: {[t.value for t in StepType]}"
            )

        references = utils.get_references_from_prompt(
            self.prompt, FLOW_REFERENCES, raise_on_invalid=True
        )

        if self.step_type == StepType.DEFAULT_STEP and (
            references.get("global_functions") or references.get("transition_functions")
        ):
            function_references = []
            if references.get("global_functions"):
                function_references.extend(references.get("global_functions").keys())
            if references.get("transition_functions"):
                function_references.extend(references.get("transition_functions").keys())
            raise ValueError(
                f"Default steps cannot reference functions. "
                f"Found function references: {function_references}"
            )

        valid, invalid_references = utils.validate_references(
            references, resource_mappings, flow_name=self.flow_name
        )
        if not valid:
            raise ValueError(f"Invalid references: {invalid_references}")

        for condition in self.conditions:
            try:
                condition.validate(resource_mappings=resource_mappings)
            except Exception as e:
                raise ValueError(f"Condition '{condition.name}': {e}") from e

        entity_ids = set(
            resource.resource_id
            for resource in resource_mappings or []
            if resource.resource_type == Entity
        )
        if self.extracted_entities:
            for entity_id in self.extracted_entities:
                if entity_id not in entity_ids:
                    raise ValueError(f"Requested entity '{entity_id}' not found.")

    def build_update_proto(
        self,
    ) -> Flow_UpdateStep | UpdateNoCodeStep:
        """Create a proto for updating the resource."""
        if self.step_type == StepType.ADVANCED_STEP:
            references = utils.get_references_from_prompt(self.prompt, FLOW_REFERENCES)
            return Flow_UpdateStep(
                flow_id=self.flow_id,
                step=UpdateAdvancedStep(
                    id=self.step_id,
                    name=self.name,
                    prompt=self.prompt,
                    references=references,
                ),
            )

        if self.step_type == StepType.DEFAULT_STEP:
            references = utils.get_references_from_prompt(self.prompt, NO_CODE_STEP_REFERENCES)
            references["extracted_entities"] = {
                entity_name: True for entity_name in self.extracted_entities
            }
            return UpdateNoCodeStep(
                flow_id=self.flow_id,
                step_id=self.step_id,
                name=self.name,
                prompt=self.prompt,
                references=references,
            )

        raise NotImplementedError("Step type not implemented")

    def build_delete_proto(self) -> DeleteNoCodeStep | DeleteStep:
        """Create a proto for deleting the resource."""
        if self.step_type == StepType.ADVANCED_STEP:
            return Flow_DeleteStep(
                flow_id=self.flow_id,
                step_id=self.step_id,
            )

        if self.step_type == StepType.DEFAULT_STEP:
            return DeleteNoCodeStep(
                flow_id=self.flow_id,
                step_id=self.step_id,
            )

        raise NotImplementedError

    def build_create_proto(
        self,
    ) -> Flow_CreateStep | CreateNoCodeStep:
        """Create a proto for creating the resource."""
        if self.step_type == StepType.ADVANCED_STEP:
            references = utils.get_references_from_prompt(self.prompt, FLOW_REFERENCES)
            return Flow_CreateStep(
                flow_id=self.flow_id,
                step=CreateAdvancedStep(
                    id=self.step_id,
                    name=self.name,
                    prompt=self.prompt,
                    references=references,
                    position=StepPosition(
                        x=self.position.get("x", 0.0), y=self.position.get("y", 0.0)
                    ),
                    asr_biasing=self.asr_biasing.build_create_proto(),
                    dtmf_config=self.dtmf_config.build_create_proto(),
                ),
            )

        if self.step_type == StepType.DEFAULT_STEP:
            references = utils.get_references_from_prompt(self.prompt, NO_CODE_STEP_REFERENCES)
            references["extracted_entities"] = {
                entity_name: True for entity_name in self.extracted_entities
            }
            return CreateNoCodeStep(
                flow_id=self.flow_id,
                step_id=self.step_id,
                name=self.name,
                prompt=self.prompt,
                position=StepPosition(x=self.position.get("x", 0.0), y=self.position.get("y", 0.0)),
                references=references,
            )

        raise NotImplementedError("Step type not implemented")

    @property
    def command_type(self) -> str:
        """Get the update type for updating the resource."""
        if self.step_type == StepType.ADVANCED_STEP:
            return "flow_step"
        if self.step_type == StepType.DEFAULT_STEP:
            return "no_code_step"

        raise NotImplementedError("Step type not implemented")

    @staticmethod
    def discover_resources(base_path: str) -> list[str]:
        """Discover resources of this type in the given base path.

        Args:
            base_path (str): The base path to search for resources.

        Returns:
            list[str]: A list of file paths of discovered resources.
        """
        flows_path = os.path.join(base_path, "flows")
        discovered_flow_steps = []

        if not os.path.exists(flows_path):
            return discovered_flow_steps

        # Find all the flow steps that match the pattern flows/<flow_name>/steps/<step_name>.yaml
        for flow_name in os.listdir(flows_path):
            steps_path = os.path.join(flows_path, flow_name, "steps")
            if os.path.exists(steps_path):
                for file_name in os.listdir(steps_path):
                    if file_name.endswith(".yaml"):
                        discovered_flow_steps.append(os.path.join(steps_path, file_name))

        return discovered_flow_steps

    def get_new_updated_deleted_subresources(
        self, old_resource: Optional["FlowStep"] = None
    ) -> tuple[list[SubResource], list[SubResource], list[SubResource]]:
        """Get the new, updated, and deleted subresources within this resource.

        Returns:
            tuple[
                list[SubResource],
                list[SubResource],
                list[SubResource],
            ]: A tuple containing three lists of subresources:
                - New subresources
                - Updated subresources
                - Deleted subresources
        """
        new = []
        updated = []
        deleted = []
        if self.step_type == StepType.ADVANCED_STEP and old_resource:
            if (
                self.asr_biasing
                and old_resource
                and (not old_resource.asr_biasing or self.asr_biasing != old_resource.asr_biasing)
            ):
                updated.append(self.asr_biasing)
            if self.dtmf_config and (
                not old_resource.dtmf_config or self.dtmf_config != old_resource.dtmf_config
            ):
                updated.append(self.dtmf_config)

        if self.step_type == StepType.DEFAULT_STEP:
            old_condition_ids = (
                {cond.resource_id for cond in old_resource.conditions} if old_resource else set()
            )
            new_condition_ids = {cond.resource_id for cond in self.conditions}

            for condition in self.conditions:
                if condition.resource_id not in old_condition_ids:
                    new.append(condition)
                else:
                    # Check if updated
                    old_condition = next(
                        (
                            c
                            for c in old_resource.conditions
                            if c.resource_id == condition.resource_id
                        ),
                        None,
                    )
                    if old_condition and condition != old_condition:
                        updated.append(condition)

            if old_resource:
                for condition in old_resource.conditions:
                    if condition.resource_id not in new_condition_ids:
                        deleted.append(condition)

        return new, updated, deleted


@dataclass
class ASRBiasing(SubResource):
    """ASR Biasing configuration."""

    alphanumeric: bool
    name_spelling: bool
    numeric: bool
    party_size: bool
    precise_date: bool
    relative_date: bool
    single_number: bool
    time: bool
    yes_no: bool
    address: bool
    custom_keywords: list[str]
    is_enabled: bool
    step_id: str
    flow_id: str

    def __init__(
        self,
        step_id: str,
        flow_id: str,
        alphanumeric: bool = False,
        name_spelling: bool = False,
        numeric: bool = False,
        party_size: bool = False,
        precise_date: bool = False,
        relative_date: bool = False,
        single_number: bool = False,
        time: bool = False,
        yes_no: bool = False,
        address: bool = False,
        custom_keywords: list[str] | None = None,
        is_enabled: bool = False,
    ):
        self.name = "asr"
        self.step_id = step_id
        self.flow_id = flow_id
        self.resource_id = f"{flow_id}.{step_id}"
        self.alphanumeric = alphanumeric
        self.name_spelling = name_spelling
        self.numeric = numeric
        self.party_size = party_size
        self.precise_date = precise_date
        self.relative_date = relative_date
        self.single_number = single_number
        self.time = time
        self.yes_no = yes_no
        self.address = address
        self.custom_keywords = custom_keywords or []
        self.is_enabled = is_enabled

    @property
    def command_type(self) -> str:
        """Get the update type for updating the resource."""
        return "flow_step_asr_config"

    def to_yaml_dict(self) -> dict:
        """Return a dictionary suitable for YAML serialization."""
        return {
            "is_enabled": self.is_enabled,
            "alphanumeric": self.alphanumeric,
            "name_spelling": self.name_spelling,
            "numeric": self.numeric,
            "party_size": self.party_size,
            "precise_date": self.precise_date,
            "relative_date": self.relative_date,
            "single_number": self.single_number,
            "time": self.time,
            "yes_no": self.yes_no,
            "address": self.address,
            "custom_keywords": self.custom_keywords,
        }

    def build_update_proto(self) -> Flow_UpdateStepAsrConfig:
        """Create a proto for updating the ASR biasing configuration."""
        return Flow_UpdateStepAsrConfig(
            flow_id=self.flow_id,
            step_id=self.step_id,
            asr_biasing=StepAsrConfigUpdate(
                alphanumeric=self.alphanumeric,
                name_spelling=self.name_spelling,
                numeric=self.numeric,
                party_size=self.party_size,
                precise_date=self.precise_date,
                relative_date=self.relative_date,
                single_number=self.single_number,
                time=self.time,
                yes_no=self.yes_no,
                address=self.address,
                custom_keywords=UpdateAsrKeywords(
                    custom_keywords=self.custom_keywords,
                ),
                is_enabled=self.is_enabled,
            ),
        )

    def build_delete_proto(self):
        """Create a proto for deleting the ASR biasing configuration."""
        raise NotImplementedError("ASR Biasing config does not support deletion.")

    def build_create_proto(self) -> StepAsrConfig:
        """Create a proto for creating the ASR biasing configuration."""
        return StepAsrConfig(
            alphanumeric=self.alphanumeric,
            name_spelling=self.name_spelling,
            numeric=self.numeric,
            party_size=self.party_size,
            precise_date=self.precise_date,
            relative_date=self.relative_date,
            single_number=self.single_number,
            time=self.time,
            yes_no=self.yes_no,
            address=self.address,
            custom_keywords=self.custom_keywords,
            is_enabled=self.is_enabled,
        )


@dataclass
class DTMFConfig(SubResource):
    """DTMF Configuration."""

    is_enabled: bool
    inter_digit_timeout: int
    max_digits: int
    end_key: str
    collect_while_agent_speaking: bool
    is_pii: bool
    step_id: str
    flow_id: str

    def __init__(
        self,
        step_id: str,
        flow_id: str,
        is_enabled: bool = False,
        inter_digit_timeout: int = 0,
        max_digits: int = 0,
        end_key: str = "#",
        collect_while_agent_speaking: bool = False,
        is_pii: bool = False,
    ):
        self.name = "dtmf"
        self.step_id = step_id
        self.flow_id = flow_id
        self.resource_id = f"{flow_id}.{step_id}"
        self.is_enabled = is_enabled
        self.inter_digit_timeout = inter_digit_timeout
        self.max_digits = max_digits
        self.end_key = end_key
        self.collect_while_agent_speaking = collect_while_agent_speaking
        self.is_pii = is_pii

    @property
    def command_type(self) -> str:
        """Get the update type for updating the resource."""
        return "flow_step_dtmf_config"

    def to_yaml_dict(self) -> dict:
        """Return a dictionary suitable for YAML serialization."""
        return {
            "is_enabled": self.is_enabled,
            "inter_digit_timeout": self.inter_digit_timeout,
            "max_digits": self.max_digits,
            "end_key": self.end_key,
            "collect_while_agent_speaking": self.collect_while_agent_speaking,
            "is_pii": self.is_pii,
        }

    def build_update_proto(self) -> Flow_UpdateStepDtmfConfig:
        """Create a proto for updating the DTMF configuration."""
        return Flow_UpdateStepDtmfConfig(
            flow_id=self.flow_id,
            step_id=self.step_id,
            dtmf_config=StepDtmfConfigUpdate(
                is_enabled=self.is_enabled,
                inter_digit_timeout=self.inter_digit_timeout,
                max_digits=self.max_digits,
                end_key=self.end_key,
                collect_while_agent_speaking=self.collect_while_agent_speaking,
                is_pii=self.is_pii,
            ),
        )

    def build_delete_proto(self):
        """Create a proto for deleting the DTMF configuration."""
        raise NotImplementedError("DTMF config deletion not implemented.")

    def build_create_proto(self) -> StepDtmfConfig:
        """Create a proto for creating the DTMF configuration."""
        return StepDtmfConfig(
            is_enabled=self.is_enabled,
            inter_digit_timeout=self.inter_digit_timeout,
            max_digits=self.max_digits,
            end_key=self.end_key,
            collect_while_agent_speaking=self.collect_while_agent_speaking,
            is_pii=self.is_pii,
        )


class ConditionType(str, Enum):
    """Enum for condition types."""

    EXIT_FLOW = "exit_flow_condition"
    STEP = "step_condition"
    NO_CODE_STEP = "no_code_step_condition"
    FUNCTION_STEP = "function_step_condition"


@dataclass
class Condition(SubResource):
    """Conditions for no code steps"""

    description: str
    required_entities: list[str]
    condition_type: ConditionType
    child_step: str
    position: dict
    exit_flow_position: dict
    ingress: str
    step_id: str
    flow_id: str

    def __init__(
        self,
        resource_id: str,
        name: str,
        condition_type: "str | ConditionType",
        step_id: str,
        flow_id: str,
        description: str = "",
        required_entities: list[str] | None = None,
        child_step: str = "",
        position: dict | None = None,
        ingress: str = "top",
        exit_flow_position: dict | None = None,
    ):
        self.resource_id = resource_id
        self.name = name
        self.description = description
        self.condition_type = (
            ConditionType(utils.to_snake_case(condition_type))
            if isinstance(condition_type, str)
            else condition_type
        )
        self.required_entities = required_entities or []
        self.child_step = child_step
        self.step_id = step_id
        self.flow_id = flow_id
        self.position = position or {}
        self.ingress = ingress
        self.exit_flow_position = exit_flow_position or {}

    def to_yaml_dict(self) -> dict:
        """Return a dictionary suitable for YAML serialization."""

        # Map ConditionType enum to YAML string
        if self.condition_type == ConditionType.EXIT_FLOW:
            condition_type = self.condition_type.value
        else:
            condition_type = "step_condition"

        yaml_dict = {
            "name": self.name,
            "condition_type": condition_type,
            "description": self.description,
        }

        if self.condition_type != ConditionType.EXIT_FLOW:
            yaml_dict["child_step"] = self.child_step

        yaml_dict["required_entities"] = self.required_entities

        return yaml_dict

    @classmethod
    def from_yaml_dict(
        cls,
        yaml_data: dict,
        resource_id: str,
        step_id: str,
        flow_id: str,
        position: dict,
        ingress: str,
        exit_flow_position: dict,
        child_step_type: Optional[StepType] = None,
    ) -> "Condition":
        """Create an instance from YAML data and identity fields."""
        if yaml_data.get("condition_type") == "step_condition":
            if child_step_type == StepType.DEFAULT_STEP:
                condition_type = ConditionType.NO_CODE_STEP
            elif child_step_type == StepType.FUNCTION_STEP:
                condition_type = ConditionType.FUNCTION_STEP
            else:
                condition_type = ConditionType.STEP
        else:
            condition_type = ConditionType.EXIT_FLOW

        return cls(
            resource_id=resource_id,
            step_id=step_id,
            flow_id=flow_id,
            position=position,
            ingress=ingress,
            exit_flow_position=exit_flow_position,
            name=yaml_data.get("name"),
            condition_type=condition_type,
            description=(yaml_data.get("description") or "").strip(),
            required_entities=yaml_data.get("required_entities", []),
            child_step=yaml_data.get("child_step", ""),
        )

    @property
    def command_type(self) -> str:
        """Get the update type for updating the resource."""
        return "no_code_condition"

    def build_update_proto(self) -> UpdateNoCodeCondition:
        """Create a proto for updating the condition."""
        return UpdateNoCodeCondition(
            flow_id=self.flow_id,
            step_id=self.step_id,
            condition_id=self.resource_id,
            **self._get_condition_type_proto(),
        )

    def build_delete_proto(self) -> DeleteNoCodeCondition:
        """Create a proto for deleting the condition."""
        return DeleteNoCodeCondition(
            flow_id=self.flow_id,
            step_id=self.step_id,
            condition_id=self.resource_id,
        )

    def build_create_proto(self) -> CreateNoCodeCondition:
        """Create a proto for creating the condition."""
        return CreateNoCodeCondition(
            flow_id=self.flow_id,
            step_id=self.step_id,
            condition_id=self.resource_id,
            **self._get_condition_type_proto(),
        )

    def _get_condition_type_proto(self) -> dict:
        """Get the condition type proto based on the condition type."""
        if self.condition_type == ConditionType.EXIT_FLOW:
            return {
                "exit_flow_condition": ExitFlowCondition(
                    details=ConditionDetails(
                        label=self.name,
                        description=self.description,
                        required_entities=self.required_entities,
                        position=self.position,
                        ingress_position=self.ingress or "top",
                    ),
                    exit_flow_position=self.exit_flow_position,
                )
            }
        elif self.condition_type == ConditionType.NO_CODE_STEP:
            return {
                "no_code_step_condition": NoCodeStepCondition(
                    details=ConditionDetails(
                        label=self.name,
                        description=self.description,
                        required_entities=self.required_entities,
                        position=self.position,
                        ingress_position=self.ingress or "top",
                    ),
                    child_step_id=self.child_step,
                )
            }
        elif self.condition_type == ConditionType.STEP:
            return {
                "step_condition": AdvancedStepCondition(
                    details=ConditionDetails(
                        label=self.name,
                        description=self.description,
                        required_entities=self.required_entities,
                        position=self.position,
                        ingress_position=self.ingress or "top",
                    ),
                    child_step_id=self.child_step,
                )
            }
        elif self.condition_type == ConditionType.FUNCTION_STEP:
            return {
                "function_step_condition": FunctionStepCondition(
                    details=ConditionDetails(
                        label=self.name,
                        description=self.description,
                        required_entities=self.required_entities,
                        position=self.position,
                        ingress_position=self.ingress or "top",
                    ),
                    child_step_id=self.child_step,
                )
            }
        else:
            raise NotImplementedError(f"Condition type {self.condition_type} not implemented.")

    def validate(self, resource_mappings: list[ResourceMapping] = None, **kwargs):
        """Validate the condition resource."""
        if not self.name:
            raise ValueError("Condition name cannot be empty.")

        if self.condition_type not in ConditionType:
            raise ValueError(f"Invalid condition type: {self.condition_type}")

        # Check child step exists in resource mappings
        # and also all required entities exist in resource mappings
        required_entity_ids = set(self.required_entities)
        if self.condition_type != ConditionType.EXIT_FLOW:
            found_step = False
        else:
            found_step = True
            # No child step to check for exit flow condition
        for resource in resource_mappings or []:
            if (
                not found_step
                and issubclass(resource.resource_type, BaseFlowStep)
                and resource.resource_id.removeprefix(resource.flow_name + "_") == self.child_step
            ):
                found_step = True

            if resource.resource_type == Entity and resource.resource_id in required_entity_ids:
                required_entity_ids.remove(resource.resource_id)

            if not required_entity_ids and found_step:
                break

        if not found_step:
            raise ValueError(f"Step '{self.child_step}' not found")

        if required_entity_ids:
            raise ValueError(f"Required entities not found: {required_entity_ids}")

        if self.description and self.description != self.description.strip():
            raise ValueError("Description cannot contain leading or trailing whitespace.")


@dataclass(init=False)
class FunctionStep(Function, BaseFlowStep):
    """Dataclass representing a function step"""

    function_id: str
    step_type: StepType = field(default=StepType.FUNCTION_STEP, init=False)
    function_type: FunctionType = field(default=FunctionType.FUNCTION_STEP, init=False)

    def __init__(
        self,
        resource_id: str,
        name: str,
        step_id: str,
        flow_id: str,
        flow_name: str,
        code: str,
        description: str = None,
        parameters: list = None,
        latency_control: dict = None,
        position: dict = None,
        function_id: str = None,
        variable_references: dict = None,
    ):
        self.step_id = step_id
        self.function_id = function_id
        self.step_type = StepType.FUNCTION_STEP
        self.position = position or {}
        super().__init__(
            resource_id=resource_id,
            name=name,
            description=None,
            code=code,
            parameters=parameters or [],
            latency_control=latency_control or {},
            flow_id=flow_id,
            flow_name=flow_name,
            function_type=FunctionType.FUNCTION_STEP,
            variable_references=variable_references,
        )

    @cached_property
    def file_path(self) -> str:
        """File path for the resource."""
        file_name = f"{self.name}.py"
        flow_name = utils.clean_name(self.flow_name)
        return os.path.join("flows", flow_name, "function_steps", file_name)

    @property
    def raw(self) -> str:
        """Convert the resource to raw format."""
        return self._generate_raw_output(
            add_description=False, add_parameters=False, add_latency_control=True
        )

    def validate(self, **kwargs):
        """Validate the resource."""
        super().validate(**kwargs)

        if self.parameters:
            raise ValueError("Function steps cannot have parameters.")

    @classmethod
    def read_local_resource(
        cls,
        file_path: str,
        resource_id: str,
        resource_name: str,
        resource_mappings: list[ResourceMapping],
        known_latency_control: dict,
        known_function_id: str = None,
        known_position: dict[str, float] = None,
        **kwargs,
    ) -> "FunctionStep":
        code = cls.read_to_raw(
            file_path, resource_mappings=resource_mappings, resource_name=resource_name, **kwargs
        )

        # Parse known latency control and extract from code (e.g. @func_latency_control)
        known_lc = Function._parse_latency_control(
            known_latency_control if known_latency_control else {}
        )
        code, _parameters, _description, latency_control = Function._extract_decorators(
            code, resource_name, [], known_lc
        )

        # e.g. flows/{flow_name}/function_steps/{function_name}.py
        parts = file_path.split(os.sep)
        if len(parts) >= 4 and parts[-4] == "flows":
            flow_folder_name = parts[-3]
        else:
            flow_folder_name = None

        flow_id = None
        flow_name = None
        if flow_folder_name:
            flow_id, flow_name = utils.get_flow_id_from_flow_name(
                flow_folder_name, resource_mappings
            )

        step_id = resource_id.removeprefix(f"{flow_name}_")

        function_id = known_function_id or f"FUNCTION-{uuid.uuid4().hex[:8]}"

        # Read references from code
        variable_references = cls._extract_variable_references(code, resource_mappings)

        return FunctionStep(
            resource_id=resource_id,
            step_id=step_id,
            name=resource_name,
            flow_id=flow_id,
            flow_name=flow_name,
            position=known_position,
            code=code,
            latency_control=latency_control,
            parameters=[],
            function_id=function_id,
            variable_references=variable_references,
        )

    @staticmethod
    def discover_resources(base_path: str) -> list[str]:
        """Discover resources of this type in the given base path.

        Args:
            base_path (str): The base path to search for resources.

        Returns:
            list[str]: A list of file paths of discovered resources.
        """
        discovered_function_steps: list[str] = []
        flows_path = os.path.join(base_path, "flows")
        if not os.path.exists(flows_path):
            return discovered_function_steps

        for flow_name in os.listdir(flows_path):
            function_steps_path = os.path.join(flows_path, flow_name, "function_steps")
            if not os.path.exists(function_steps_path):
                continue

            discovered_function_steps.extend(
                [
                    os.path.join(function_steps_path, file_name)
                    for file_name in os.listdir(function_steps_path)
                    if file_name.endswith(".py")
                ]
            )

        return discovered_function_steps

    @property
    def command_type(self) -> str:
        """Get the update type for updating the resource."""
        return "step"

    def build_update_proto(self) -> UpdateStep:
        """Create a proto for updating the resource."""
        return UpdateStep(
            flow_id=self.flow_id,
            step_id=self.step_id,
            function_step=UpdateFunctionStep(
                name=self.name,
                position=self.position,
                function=UpdateFunctionStepDefinition(
                    code=self.code,
                    latency_control=self._build_create_latency_control_proto(),
                ),
            ),
        )

    def build_delete_proto(self) -> DeleteStep:
        """Create a proto for deleting the resource."""
        return DeleteStep(
            flow_id=self.flow_id,
            step_id=self.step_id,
        )

    def build_create_proto(self) -> CreateStep:
        """Create a proto for creating the resource."""
        return CreateStep(
            flow_id=self.flow_id,
            function_step=CreateFunctionStep(
                id=self.step_id,
                name=self.name,
                position=self.position,
                function=CreateFunctionStepDefinition(
                    id=self.function_id,
                    name=self.name,
                    errors=[],
                    code=self.code,
                    latency_control=self._build_create_latency_control_proto(),
                ),
            ),
        )

    def get_new_updated_deleted_subresources(
        self, old_resource: Optional["FunctionStep"] = None
    ) -> tuple[list[SubResource], list[SubResource], list[SubResource]]:
        """LatencyControl is already included in the step update/create protos,
        so skip emitting it as a separate sub-resource command."""
        return [], [], []



"""Handling and managing an Agent Studio Function

Copyright PolyAI Limited
"""

import ast
import logging
import os
import re
import typing as ty
import uuid
from dataclasses import dataclass, field
from enum import Enum
from functools import cached_property
from typing import Literal, Optional, Union

from google.protobuf.message import Message

import poly.resources.resource_utils as utils
from poly.handlers.protobuf.end_function_pb2 import (
    EndFunction_Create,
    EndFunction_Delete,
    EndFunction_Update,
)
from poly.handlers.protobuf.flows_pb2 import (
    Flow_CreateTransitionFunction,
    Flow_DeleteTransitionFunction,
    Flow_UpdateTransitionFunction,
    Flow_UpdateTransitionFunctionLatencyControl,
    TransitionFunction_CreateTransitionFunction,
    TransitionFunction_UpdateTransitionFunction,
)
from poly.handlers.protobuf.functions_pb2 import (
    DelayResponsesUpdate,
    DelayResponseUpdate,
    Function_CreateFunction,
    Function_DeleteFunction,
    Function_UpdateFunction,
    Function_UpdateLatencyControl,
    FunctionCreateLatencyControl,
    FunctionParameterUpdate,
    ParametersUpdate,
)
from poly.handlers.protobuf.functions_pb2 import (
    FunctionDelayResponse as FunctionDelayResponseProto,
)
from poly.handlers.protobuf.start_function_pb2 import (
    StartFunction_Create,
    StartFunction_Delete,
    StartFunction_Update,
)
from poly.resources.resource import Resource, ResourceMapping, SubResource

logger = logging.getLogger(__name__)

FUNCTION_HEADER = "from _gen import *  # <AUTO GENERATED>\n"
LEGACY_FUNCTION_HEADER = "from imports import *  # <AUTO GENERATED>\n"

SchemaType = Literal["string", "integer", "number", "boolean"]

PY_TO_SCHEMA: dict[str, SchemaType] = {
    "str": "string",
    "int": "integer",
    "float": "number",
    "bool": "boolean",
}

SCHEMA_TO_PY = {v: k for k, v in PY_TO_SCHEMA.items()}


class FunctionType(str, Enum):
    """Enum representing the type of function"""

    START = "start"
    END = "end"
    TRANSITION = "transition"
    GLOBAL = "global"
    FUNCTION_STEP = "function_step"


@dataclass
class FunctionParameters:
    """Dataclass representing a function parameter"""

    name: str
    type: SchemaType
    description: str
    id: Optional[str] = None


@dataclass
class FunctionDelayResponse:
    """Dataclass representing a single delay response message"""

    message: str
    duration: int
    id: Optional[str] = None


@dataclass
class FunctionLatencyControl:
    """Dataclass representing latency control settings"""

    enabled: bool = False
    initial_delay: int = 0
    interval: int = 0
    delay_responses: list[FunctionDelayResponse] = field(default_factory=list)

    def __post_init__(self):
        self.delay_responses = [
            FunctionDelayResponse(**delay_response)
            if isinstance(delay_response, dict)
            else delay_response
            for delay_response in self.delay_responses
        ]


@dataclass
class LatencyControl(SubResource):
    """Wrapper for Latency Control protos"""

    function_id: str
    latency_control: FunctionLatencyControl
    flow_id: Optional[str] = None

    def __init__(
        self,
        *,
        function_id: str,
        latency_control: FunctionLatencyControl,
        flow_id: Optional[str] = None,
    ):
        self.name = "latency_control"
        self.resource_id = function_id
        self.function_id = function_id
        self.latency_control = latency_control
        self.flow_id = flow_id

    @property
    def command_type(self) -> str:
        return "flow_transition_function_latency_control" if self.flow_id else "latency_control"

    def to_proto(self, enabled_override: Optional[bool] = None) -> Function_UpdateLatencyControl:
        enabled = self.latency_control.enabled if enabled_override is None else enabled_override
        delay_responses = DelayResponsesUpdate(
            delay_responses=[
                DelayResponseUpdate(
                    id=dr.id or f"DELAY-{uuid.uuid4().hex[:8]}",
                    message=dr.message,
                    duration=dr.duration,
                )
                for dr in self.latency_control.delay_responses
            ]
        )
        return Function_UpdateLatencyControl(
            function_id=self.function_id,
            enabled=enabled,
            delay_responses=delay_responses,
            initial_delay=self.latency_control.initial_delay if enabled else 0,
            interval=self.latency_control.interval if enabled else 0,
        )

    def build_update_proto(self) -> Message:
        proto = self.to_proto()
        if self.flow_id:
            return Flow_UpdateTransitionFunctionLatencyControl(
                flow_id=self.flow_id, latency_control=proto
            )
        return proto

    def build_create_proto(self) -> Message:
        return self.build_update_proto()

    def build_delete_proto(self) -> Message:
        raise NotImplementedError("Latency Control does not support deletion")


@dataclass
class Function(Resource):
    """Dataclass representing an Agent Studio function"""

    description: str
    code: str
    parameters: list[FunctionParameters]
    latency_control: FunctionLatencyControl
    flow_id: Optional[str] = None
    flow_name: Optional[str] = None
    function_type: Optional[FunctionType] = None
    variable_references: Optional[dict] = None

    def __init__(
        self,
        *,
        resource_id: str,
        name: str,
        description: str,
        code: str,
        parameters: list[FunctionParameters],
        latency_control: dict | FunctionLatencyControl,
        function_type: FunctionType,
        flow_id: str | None = None,
        flow_name: str | None = None,
        variable_references: dict | None = None,
    ):
        self.resource_id = resource_id
        self.name = name
        self.description = description
        self.code = code
        self.parameters = [
            FunctionParameters(**p) if isinstance(p, dict) else p for p in (parameters or [])
        ]
        self.latency_control = self._parse_latency_control(latency_control)
        self.flow_id = flow_id
        self.flow_name = flow_name
        self.function_type = function_type
        self.variable_references = variable_references

    @staticmethod
    def _parse_latency_control(value) -> FunctionLatencyControl:
        if value is None or value == {}:
            return FunctionLatencyControl()

        if isinstance(value, FunctionLatencyControl):
            return value

        if isinstance(value, dict):
            return FunctionLatencyControl(
                enabled=value.get("enabled", False),
                initial_delay=value.get("initial_delay", value.get("initialDelay", 0)),
                interval=value.get("interval", 0),
                delay_responses=value.get("delay_responses", []),
            )
        return FunctionLatencyControl()

    @staticmethod
    def get_function_type(file_path: str) -> Optional[FunctionType]:
        """Get the function type from the file path and function name.

        Args:
            file_path (str): The file path of the function.

        Returns:
            Optional[FunctionType]: The function type.
        """
        parts = file_path.split(os.sep)
        function_name = os.path.splitext(parts[-1])[0]
        if len(parts) >= 4 and parts[-4] == "flows":
            if parts[-2] == "function_steps":
                return FunctionType.FUNCTION_STEP
            elif parts[-2] == "functions":
                return FunctionType.TRANSITION
        elif function_name == "start_function":
            return FunctionType.START
        elif function_name == "end_function":
            return FunctionType.END
        else:
            return FunctionType.GLOBAL

    @staticmethod
    def get_resource_prefix(file_path: str) -> str:
        """Get the resource prefix for the resource based on file path.

        Args:
            file_path (str): The file path of the resource.

        Returns:
            str: The resource prefix.
        """
        function_type = Function.get_function_type(file_path)
        if function_type == FunctionType.TRANSITION:
            return "ft"
        if function_type in (FunctionType.START, FunctionType.END, FunctionType.FUNCTION_STEP):
            return None
        return "fn"

    @cached_property
    def file_path(self) -> str:
        """File path for the resource."""
        file_name = f"{self.name}.py"
        if self.flow_name:
            flow_name = utils.clean_name(self.flow_name)
            return os.path.join("flows", flow_name, "functions", file_name)
        return os.path.join("functions", file_name)

    def _generate_raw_output(
        self,
        add_description: bool = True,
        add_parameters: bool = True,
        add_latency_control: bool = True,
    ) -> str:
        """Generate the pretty code for the function."""
        code = self.code

        try:
            target = self._get_target_function(code, self.name)
        except SyntaxError as e:
            logger.error(f"Syntax error while reading function {self.name!r} e={e!r}")
            return code

        has_decorators = (
            (self.parameters and add_parameters)
            or (self.description and add_description)
            or (self.latency_control and self.latency_control.enabled and add_latency_control)
        )
        if not has_decorators or not target:
            return code

        try:
            lines = code.splitlines(True)
            # Insert above the top-most decorator in the decorator block.
            insert_at = target.lineno - 1
            while insert_at > 0 and lines[insert_at - 1].lstrip().startswith("@"):
                insert_at -= 1
            indent = re.match(r"^([ \t]*)", lines[target.lineno - 1]).group(1)
            decorator_block = []
            if add_description:
                decorator_block.append(f"{indent}@func_description({self.description!r})\n")
            if add_parameters:
                decorator_block.extend(
                    f"{indent}@func_parameter({p.name!r}, {p.description!r})\n"
                    for p in self.parameters
                )
            if add_latency_control and self.latency_control and self.latency_control.enabled:
                decorator_block.append(
                    self._render_latency_control_decorator(self.latency_control, indent)
                )
            if decorator_block:
                lines.insert(insert_at, "".join(decorator_block))
            return "".join(lines)
        except SyntaxError as e:
            logger.error(f"error while converting function to raw format e={e!r}")
            return code  # return original code without injection if invalid

    @property
    def raw(self) -> str:
        """Convert the resource to raw format."""

        return self._generate_raw_output(
            add_description=True, add_parameters=True, add_latency_control=True
        )

    @staticmethod
    def make_pretty(contents: str, resource_mappings: list[ResourceMapping], **kwargs) -> str:
        """Format the raw representation of the resource."""
        # Removing leading newlines
        contents = contents.lstrip("\n")

        # Check if contents starts with a module-level docstring
        docstring_end = None
        if contents.startswith('"""') or contents.startswith("'''"):
            quote_type = contents[:3]
            # Find the closing quote
            end_pos = contents.find(quote_type, 3)
            if end_pos != -1:
                docstring_end = end_pos + 3

        # If there's a docstring, insert FUNCTION_HEADER after it
        if docstring_end is not None:
            before_docstring = contents[:docstring_end]
            after_docstring = contents[docstring_end:].lstrip("\n")

            # Add header after docstring with appropriate spacing
            if after_docstring and (
                after_docstring.startswith("from ") or after_docstring.startswith("import ")
            ):
                # There are imports right after, add header with single newline
                code = before_docstring + "\n" + FUNCTION_HEADER + after_docstring
            else:
                # No imports right after, add header with two newlines
                code = before_docstring + "\n" + FUNCTION_HEADER + "\n" + after_docstring
        else:
            # No docstring, add header at the top
            # If there are no imports, add two newlines after the header
            if not contents.startswith(("from ", "import ")):
                code = FUNCTION_HEADER + "\n\n" + contents
            else:
                code = FUNCTION_HEADER + contents

        # Convert `functions.{flow_id}` to `flows.{flow_name}.functions`
        for resource in resource_mappings:
            if resource.resource_type.__name__ == "FlowConfig":
                code = code.replace(
                    f"functions.{utils.clean_name(resource.resource_id)}",
                    f"flows.{utils.clean_name(resource.resource_name)}.functions",
                )

        return code

    @staticmethod
    def from_pretty(
        contents: str, resource_mappings: list[ResourceMapping], resource_name: str = None, **kwargs
    ) -> str:
        """Undo formatting or changes made to the local resource."""
        # Remove typing import if it exists
        code = contents

        if not code:
            return code

        if FUNCTION_HEADER in code:
            code = code.replace(FUNCTION_HEADER, "")
        if LEGACY_FUNCTION_HEADER in code:
            code = code.replace(LEGACY_FUNCTION_HEADER, "")

        # Remove leading newlines
        code = code.lstrip("\n")

        # Convert `flows.{flow_name}.functions` to `functions.{flow_id}`
        for resource in resource_mappings:
            if resource.resource_type.__name__ == "FlowConfig":
                code = code.replace(
                    f"flows.{utils.clean_name(resource.resource_name)}.functions",
                    f"functions.{utils.clean_name(resource.resource_id)}",
                )

        code = utils.restore_function_def_line(code, resource_name)

        return code

    @staticmethod
    def _param_def_string(parameter: FunctionParameters):
        """Get the def string for the parameter"""
        return f"{parameter.name}: {SCHEMA_TO_PY.get(parameter.type, 'str')}"

    @staticmethod
    def _render_latency_control_decorator(lc: FunctionLatencyControl, indent: str) -> str:
        parts = [
            f"delay_before_responses_start={lc.initial_delay!r}",
            f"silence_after_each_response={lc.interval!r}",
        ]
        if lc.delay_responses:
            dr_items = ", ".join(f"({dr.message!r}, {dr.duration!r})" for dr in lc.delay_responses)
            parts.append(f"delay_responses=[{dr_items}]")
        return f"{indent}@func_latency_control({', '.join(parts)})\n"

    def validate(self, **kwargs) -> None:
        """Validate the resource.

        Raises:
            ValidationError: If the resource is not valid.
        """
        # Check for syntax errors
        try:
            compile(self.code, self.name, "exec")
        except SyntaxError as e:
            logger.debug(f"Syntax error validating code={self.code!r}")
            raise ValueError(f"Syntax error in function code: {e}")

        # Check function def exists in code
        default_params = ["conv: Conversation"]
        if self.function_type in (FunctionType.TRANSITION, FunctionType.FUNCTION_STEP):
            default_params.append("flow: Flow")
        params = default_params + [self._param_def_string(p) for p in self.parameters]
        function_def = f"def {self.name}({', '.join(params)})"

        if function_def not in self.code:
            raise ValueError(f"Function definition '{function_def}' not found in code.")

        # Ensure description exists
        if not self.description and self.function_type not in [
            FunctionType.START,
            FunctionType.END,
            FunctionType.FUNCTION_STEP,
        ]:
            raise ValueError("Function description cannot be empty.")

        if self.description and self.description != self.description.strip():
            raise ValueError("Description cannot contain leading or trailing whitespace.")

        # Ensure flow_id is set for transition functions
        if self.function_type == FunctionType.TRANSITION and not self.flow_id:
            raise ValueError("Can't find flow_id for transition function.")

        # Validate latency control settings
        if self.latency_control and self.latency_control.enabled:
            if not (0 <= self.latency_control.initial_delay <= 10):
                raise ValueError(
                    "delay_before_responses_start must be between 0 and 10, "
                    f"got {self.latency_control.initial_delay}."
                )
            if not (0 <= self.latency_control.interval <= 10):
                raise ValueError(
                    "silence_after_each_response must be between 0 and 10, "
                    f"got {self.latency_control.interval}."
                )
            if not self.latency_control.delay_responses:
                raise ValueError("delay_responses cannot be empty.")

        for line in self.code.splitlines():
            if line.strip().startswith("#"):
                continue
            if (
                line.startswith("@func_description")
                or line.startswith("@func_parameter")
                or line.startswith("@func_latency_control")
            ):
                raise ValueError(
                    "ADK decorators found in raw code. This might be because of a parameter mismatch."
                )

        resource_mappings = kwargs.get("resource_mappings") or []
        code_for_validation = utils.remove_comments_from_code(self.code)

        if self.flow_name and resource_mappings:
            valid_step_names = {
                r.resource_name
                for r in resource_mappings
                if r.resource_type.__name__ in ("FlowStep", "FunctionStep")
                and r.flow_name == self.flow_name
            }
            if valid_step_names:
                for match in re.finditer(
                    r'flow\.goto_step\(\s*"((?:[^"\\]|\\.)*)"'
                    r"|flow\.goto_step\(\s*'((?:[^'\\]|\\.)*)'",
                    code_for_validation,
                ):
                    target_step = match.group(1) or match.group(2)
                    if target_step not in valid_step_names:
                        raise ValueError(
                            f"flow.goto_step('{target_step}') references a step that does not exist "
                            f"in flow '{self.flow_name}'."
                        )

        if resource_mappings:
            valid_flow_names = {
                r.resource_name
                for r in resource_mappings
                if r.resource_type.__name__ == "FlowConfig"
            }
            if valid_flow_names:
                for match in re.finditer(
                    r'conv\.goto_flow\(\s*"((?:[^"\\]|\\.)*)"'
                    r"|conv\.goto_flow\(\s*'((?:[^'\\]|\\.)*)'",
                    code_for_validation,
                ):
                    target_flow = match.group(1) or match.group(2)
                    if target_flow not in valid_flow_names:
                        raise ValueError(
                            f"conv.goto_flow('{target_flow}') references a flow that does not exist."
                        )

    @staticmethod
    def _get_target_function(
        code: str, function_name: str
    ) -> Optional[Union[ast.FunctionDef, ast.AsyncFunctionDef]]:
        module = ast.parse(code)
        return next(
            (
                f
                for f in ast.walk(module)
                if isinstance(f, (ast.FunctionDef, ast.AsyncFunctionDef))
                and f.name == function_name
            ),
            None,
        )

    @staticmethod
    def _extract_decorators(
        code: str,
        function_name: str,
        known_parameters: list[FunctionParameters],
        known_latency_control: Optional[FunctionLatencyControl] = None,
    ) -> tuple[str, list[FunctionParameters], Optional[str], FunctionLatencyControl]:
        """Extract decorators from the function code.

        Args:
            code (str): The function code.

        Returns:
            tuple: The cleaned code, list of parameters, and description.
        """
        parameters: list[FunctionParameters] = []
        description: str = None
        latency_control = FunctionLatencyControl()
        try:
            target = Function._get_target_function(code, function_name)
            if target:
                removable_lines: set[int] = set()
                for decorator in target.decorator_list:
                    if not (hasattr(decorator, "func") and hasattr(decorator.func, "id")):
                        continue

                    decorator_name = decorator.func.id

                    if decorator_name == "func_parameter" and len(decorator.args) == 2:
                        name, desc = (arg.value for arg in decorator.args)
                        _type = next(
                            (arg.annotation.id for arg in target.args.args if arg.arg == name),
                            None,
                        )
                        if not _type:
                            raise SyntaxError(f"Parameter {name!r} is missing a type.")

                        _id = next(
                            (param.id for param in known_parameters if param.name == name),
                            f"PARAMETER-{uuid.uuid4().hex[:8]}",
                        )

                        if _type in PY_TO_SCHEMA:
                            parameters.append(
                                FunctionParameters(
                                    id=_id,
                                    name=name,
                                    description=desc,
                                    type=PY_TO_SCHEMA[_type],
                                )
                            )
                            removable_lines.update(
                                range(decorator.lineno - 1, decorator.end_lineno)
                            )
                    elif decorator_name == "func_description" and len(decorator.args) == 1:
                        description = (decorator.args[0].value).strip()
                        removable_lines.update(range(decorator.lineno - 1, decorator.end_lineno))

                    elif decorator_name == "func_latency_control":
                        latency_control = Function._parse_latency_control_decorator(
                            decorator, known_latency_control
                        )
                        removable_lines.update(range(decorator.lineno - 1, decorator.end_lineno))

                if removable_lines:
                    lines = code.splitlines(True)
                    code = "".join(
                        line for idx, line in enumerate(lines) if idx not in removable_lines
                    )
        except SyntaxError as e:
            logger.error(f"error while extracting decorators from function e={e!r}")
        return code, parameters, description, latency_control

    @staticmethod
    def _parse_latency_control_decorator(
        decorator: ast.Call,
        known_latency_control: Optional[FunctionLatencyControl] = None,
    ) -> FunctionLatencyControl:
        """Parse @func_latency_control into FunctionLatencyControl dataclass."""

        initial_delay = 0
        interval = 0
        delay_responses: list[FunctionDelayResponse] = []
        used_delay_response_ids: set[str] = set()

        for kw in decorator.keywords:
            if kw.arg == "delay_before_responses_start" and isinstance(kw.value, ast.Constant):
                initial_delay = kw.value.value
            elif kw.arg == "silence_after_each_response" and isinstance(kw.value, ast.Constant):
                interval = kw.value.value
            elif kw.arg == "delay_responses" and isinstance(kw.value, ast.List):
                for elt in kw.value.elts:
                    if isinstance(elt, ast.Tuple) and len(elt.elts) == 2:
                        msg = elt.elts[0].value if isinstance(elt.elts[0], ast.Constant) else ""
                        dur = elt.elts[1].value if isinstance(elt.elts[1], ast.Constant) else 0
                        # Preserve existing ID when the message text matches.
                        existing_id = None
                        if known_latency_control and known_latency_control.delay_responses:
                            existing_id = next(
                                (
                                    dr.id
                                    for dr in known_latency_control.delay_responses
                                    if dr.message == msg and dr.id not in used_delay_response_ids
                                ),
                                None,
                            )
                        if existing_id:
                            used_delay_response_ids.add(existing_id)
                        delay_responses.append(
                            FunctionDelayResponse(
                                id=existing_id or f"DELAY-{uuid.uuid4().hex[:8]}",
                                message=msg,
                                duration=dur,
                            )
                        )

        return FunctionLatencyControl(
            enabled=True,  # if decorator exists, implies enabled = True
            initial_delay=initial_delay,
            interval=interval,
            delay_responses=delay_responses,
        )

    @staticmethod
    def _extract_variable_references(code: str, resource_mappings: list[ResourceMapping]) -> dict:
        """Extract variable references from code."""
        variable_references = {}
        variable_names = utils.extract_variable_names_from_code(code)

        if variable_names:
            # Swap variable names for ids:
            known_variables = {
                v.resource_name: v.resource_id
                for v in resource_mappings
                if v.resource_type.__name__ == "Variable"
            }
            for name in variable_names:
                if name not in known_variables:
                    logger.warning(
                        f"Variable {name} not found in resource mappings, will be added in the next push"
                    )
                    continue
                variable_references[known_variables[name]] = True
        return variable_references

    @classmethod
    def read_local_resource(
        cls,
        file_path: str,
        resource_id: str,
        resource_name: str,
        resource_mappings: list[ResourceMapping],
        known_parameters: list[FunctionParameters],
        known_latency_control: FunctionLatencyControl = None,
        **kwargs,
    ) -> "Function":
        """Read a local YAML resource from the given file path."""

        code = cls.read_to_raw(
            file_path, resource_mappings=resource_mappings, resource_name=resource_name, **kwargs
        )

        if known_latency_control is None:
            known_latency_control = FunctionLatencyControl()

        # Extract description, parameters, latency_control from code
        code, parameters, description, latency_control = cls._extract_decorators(
            code, resource_name, known_parameters, known_latency_control
        )

        # flow_name can be inferred from file path
        # e.g. flows/{flow_name}/functions/{function_name}.py
        parts = file_path.split(os.sep)
        if len(parts) >= 4 and parts[-4] == "flows":
            flow_folder_name = parts[-3]
        else:
            flow_folder_name = None

        # Search for flow_id in resource mappings
        flow_id = None
        flow_name = None
        if flow_folder_name:
            flow_id, flow_name = utils.get_flow_id_from_flow_name(
                flow_folder_name, resource_mappings
            )

        # function_type can be inferred from file path
        function_type = cls.get_function_type(file_path)

        # Read references from code
        variable_references = cls._extract_variable_references(code, resource_mappings)

        return Function(
            resource_id=resource_id,
            name=resource_name,
            description=description,
            code=code,
            parameters=parameters,
            latency_control=latency_control,
            flow_id=flow_id,
            flow_name=flow_name,
            function_type=function_type,
            variable_references=variable_references,
        )

    @staticmethod
    def _create_function_param_updates(
        parameters: list[FunctionParameters],
    ) -> FunctionParameterUpdate:
        """Create parameter updates for the function."""
        return [
            FunctionParameterUpdate(
                id=p.id,
                name=p.name,
                description=p.description,
                type=p.type,
            )
            for p in parameters
        ]

    def build_update_proto(
        self,
    ) -> ty.Union[
        StartFunction_Update,
        EndFunction_Update,
        Flow_UpdateTransitionFunction,
        Function_UpdateFunction,
    ]:
        """Create a proto for updating the resource."""

        if self.parameters:
            params_update = ParametersUpdate(
                parameters=self._create_function_param_updates(self.parameters)
            )
        else:
            params_update = None

        fn_proto = {
            "id": self.resource_id,
            "code": self.code,
            "description": self.description,
        }
        if params_update is not None:
            fn_proto["parameters"] = params_update

        if self.function_type != FunctionType.TRANSITION:
            fn_proto["references"] = {
                "variables": self.variable_references or {},
            }

        if self.function_type == FunctionType.START:
            return StartFunction_Update(**fn_proto)
        elif self.function_type == FunctionType.END:
            return EndFunction_Update(**fn_proto)
        elif self.function_type == FunctionType.TRANSITION:
            return Flow_UpdateTransitionFunction(
                flow_id=self.flow_id,
                transition_function=TransitionFunction_UpdateTransitionFunction(**fn_proto),
            )

        return Function_UpdateFunction(**fn_proto)

    def _build_create_latency_control_proto(self) -> FunctionCreateLatencyControl:
        delay_responses = [
            FunctionDelayResponseProto(
                id=dr.id or f"DELAY-{uuid.uuid4().hex[:8]}",
                message=dr.message,
                duration=dr.duration,
            )
            for dr in self.latency_control.delay_responses
        ]
        return FunctionCreateLatencyControl(
            enabled=self.latency_control.enabled,
            delay_responses=delay_responses,
            initial_delay=self.latency_control.initial_delay,
            interval=self.latency_control.interval,
        )

    def build_create_proto(self) -> Message:
        """Create a proto for creating the resource."""
        if self.parameters:
            params_update = self._create_function_param_updates(self.parameters)
        else:
            params_update = None

        fn_proto = {
            "id": self.resource_id,
            "code": self.code,
            "name": self.name,
            "description": self.description,
        }
        if params_update is not None:
            fn_proto["parameters"] = params_update

        fn_proto["references"] = {
            "variables": self.variable_references or {},
        }

        if self.function_type == FunctionType.START:
            return StartFunction_Create(**fn_proto)
        elif self.function_type == FunctionType.END:
            return EndFunction_Create(**fn_proto)
        elif self.function_type == FunctionType.TRANSITION:
            fn_proto["references"]["flow_steps"] = {}
            return Flow_CreateTransitionFunction(
                flow_id=self.flow_id,
                transition_function=TransitionFunction_CreateTransitionFunction(**fn_proto),
            )

        if self.latency_control and self.latency_control.enabled:
            fn_proto["latency_control"] = self._build_create_latency_control_proto()

        return Function_CreateFunction(**fn_proto)

    def build_delete_proto(self) -> Message:
        """Create a proto for deleting the resource."""
        if self.function_type == FunctionType.START:
            return StartFunction_Delete(id=self.resource_id)
        elif self.function_type == FunctionType.END:
            return EndFunction_Delete(id=self.resource_id)
        elif self.function_type == FunctionType.TRANSITION:
            return Flow_DeleteTransitionFunction(
                flow_id=self.flow_id,
                function_id=self.resource_id,
            )
        return Function_DeleteFunction(id=self.resource_id)

    def get_new_updated_deleted_subresources(
        self, old_resource: ty.Optional["Function"] = None
    ) -> tuple[list[SubResource], list[SubResource], list[SubResource]]:
        """Detect latency-control changes that need their own command."""
        new: list[SubResource] = []
        updated: list[SubResource] = []

        new_lc = self.latency_control
        old_lc = old_resource.latency_control if old_resource else FunctionLatencyControl()

        if new_lc != old_lc:
            sub = LatencyControl(
                function_id=self.resource_id,
                latency_control=new_lc,
                flow_id=self.flow_id if self.function_type == FunctionType.TRANSITION else None,
            )
            updated.append(sub)

        return new, updated, []

    @property
    def command_type(self) -> str:
        """Get the update type for updating the resource."""
        if self.function_type == FunctionType.START:
            return "start_function"
        elif self.function_type == FunctionType.END:
            return "end_function"
        elif self.function_type == FunctionType.TRANSITION:
            return "flow_transition_function"
        elif self.function_type == FunctionType.GLOBAL:
            return "function"

    @staticmethod
    def discover_resources(base_path: str) -> list[str]:
        """Discover resources of this type in the given base path.

        Args:
            base_path (str): The base path to search for resources.

        Returns:
            list[str]: A list of file paths of discovered resources.
        """
        discovered_functions: list[str] = []

        # Find all transition functions in flows/*/functions/
        functions_path = os.path.join(base_path, "flows")
        discovered_flow_functions: list[str] = []
        if os.path.exists(functions_path):
            discovered_flow_functions = os.listdir(functions_path)
        for flow_name in discovered_flow_functions:
            flow_functions_path = os.path.join(functions_path, flow_name, "functions")
            if not os.path.exists(flow_functions_path):
                continue

            discovered_functions.extend(
                [
                    os.path.join(flow_functions_path, file_name)
                    for file_name in os.listdir(flow_functions_path)
                    if file_name.endswith(".py")
                ]
            )

        # Find all global functions in functions/
        functions_path = os.path.join(base_path, "functions")
        if not os.path.exists(functions_path):
            return discovered_functions

        discovered_functions.extend(
            [
                os.path.join(functions_path, file_name)
                for file_name in os.listdir(functions_path)
                if file_name.endswith(".py")
            ]
        )

        return discovered_functions

    @staticmethod
    def format_resource(content, file_name: str) -> str:
        """Format the resource code using ruff."""
        content = utils.format_code_with_ruff(content, file_name)
        content = utils.restore_function_def_line(content, file_name)
        return content



from google.protobuf import timestamp_pb2 as _timestamp_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class ExampleQueries(_message.Message):
    __slots__ = ("queries",)
    QUERIES_FIELD_NUMBER: _ClassVar[int]
    queries: _containers.RepeatedScalarFieldContainer[str]
    def __init__(self, queries: _Optional[_Iterable[str]] = ...) -> None: ...

class TopicReferences(_message.Message):
    __slots__ = ("sms", "handoff", "attributes", "global_functions", "variables")
    class SmsEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: bool
        def __init__(self, key: _Optional[str] = ..., value: bool = ...) -> None: ...
    class HandoffEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: bool
        def __init__(self, key: _Optional[str] = ..., value: bool = ...) -> None: ...
    class AttributesEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: bool
        def __init__(self, key: _Optional[str] = ..., value: bool = ...) -> None: ...
    class GlobalFunctionsEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: bool
        def __init__(self, key: _Optional[str] = ..., value: bool = ...) -> None: ...
    class VariablesEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: bool
        def __init__(self, key: _Optional[str] = ..., value: bool = ...) -> None: ...
    SMS_FIELD_NUMBER: _ClassVar[int]
    HANDOFF_FIELD_NUMBER: _ClassVar[int]
    ATTRIBUTES_FIELD_NUMBER: _ClassVar[int]
    GLOBAL_FUNCTIONS_FIELD_NUMBER: _ClassVar[int]
    VARIABLES_FIELD_NUMBER: _ClassVar[int]
    sms: _containers.ScalarMap[str, bool]
    handoff: _containers.ScalarMap[str, bool]
    attributes: _containers.ScalarMap[str, bool]
    global_functions: _containers.ScalarMap[str, bool]
    variables: _containers.ScalarMap[str, bool]
    def __init__(self, sms: _Optional[_Mapping[str, bool]] = ..., handoff: _Optional[_Mapping[str, bool]] = ..., attributes: _Optional[_Mapping[str, bool]] = ..., global_functions: _Optional[_Mapping[str, bool]] = ..., variables: _Optional[_Mapping[str, bool]] = ...) -> None: ...

class KnowledgeBase(_message.Message):
    __slots__ = ("topics", "embedding_model")
    TOPICS_FIELD_NUMBER: _ClassVar[int]
    EMBEDDING_MODEL_FIELD_NUMBER: _ClassVar[int]
    topics: _containers.RepeatedCompositeFieldContainer[KnowledgeBaseTopic]
    embedding_model: str
    def __init__(self, topics: _Optional[_Iterable[_Union[KnowledgeBaseTopic, _Mapping]]] = ..., embedding_model: _Optional[str] = ...) -> None: ...

class KnowledgeBaseTopic(_message.Message):
    __slots__ = ("id", "name", "content", "actions", "example_queries", "references", "created_at", "created_by", "updated_at", "updated_by", "is_active")
    ID_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    CONTENT_FIELD_NUMBER: _ClassVar[int]
    ACTIONS_FIELD_NUMBER: _ClassVar[int]
    EXAMPLE_QUERIES_FIELD_NUMBER: _ClassVar[int]
    REFERENCES_FIELD_NUMBER: _ClassVar[int]
    CREATED_AT_FIELD_NUMBER: _ClassVar[int]
    CREATED_BY_FIELD_NUMBER: _ClassVar[int]
    UPDATED_AT_FIELD_NUMBER: _ClassVar[int]
    UPDATED_BY_FIELD_NUMBER: _ClassVar[int]
    IS_ACTIVE_FIELD_NUMBER: _ClassVar[int]
    id: str
    name: str
    content: str
    actions: str
    example_queries: ExampleQueries
    references: TopicReferences
    created_at: _timestamp_pb2.Timestamp
    created_by: str
    updated_at: _timestamp_pb2.Timestamp
    updated_by: str
    is_active: bool
    def __init__(self, id: _Optional[str] = ..., name: _Optional[str] = ..., content: _Optional[str] = ..., actions: _Optional[str] = ..., example_queries: _Optional[_Union[ExampleQueries, _Mapping]] = ..., references: _Optional[_Union[TopicReferences, _Mapping]] = ..., created_at: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., created_by: _Optional[str] = ..., updated_at: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., updated_by: _Optional[str] = ..., is_active: bool = ...) -> None: ...

class KnowledgeBase_CreateTopic(_message.Message):
    __slots__ = ("id", "name", "content", "actions", "example_queries", "references", "is_active")
    ID_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    CONTENT_FIELD_NUMBER: _ClassVar[int]
    ACTIONS_FIELD_NUMBER: _ClassVar[int]
    EXAMPLE_QUERIES_FIELD_NUMBER: _ClassVar[int]
    REFERENCES_FIELD_NUMBER: _ClassVar[int]
    IS_ACTIVE_FIELD_NUMBER: _ClassVar[int]
    id: str
    name: str
    content: str
    actions: str
    example_queries: ExampleQueries
    references: TopicReferences
    is_active: bool
    def __init__(self, id: _Optional[str] = ..., name: _Optional[str] = ..., content: _Optional[str] = ..., actions: _Optional[str] = ..., example_queries: _Optional[_Union[ExampleQueries, _Mapping]] = ..., references: _Optional[_Union[TopicReferences, _Mapping]] = ..., is_active: bool = ...) -> None: ...

class KnowledgeBase_UpdateTopic(_message.Message):
    __slots__ = ("id", "name", "content", "actions", "example_queries", "references", "is_active")
    ID_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    CONTENT_FIELD_NUMBER: _ClassVar[int]
    ACTIONS_FIELD_NUMBER: _ClassVar[int]
    EXAMPLE_QUERIES_FIELD_NUMBER: _ClassVar[int]
    REFERENCES_FIELD_NUMBER: _ClassVar[int]
    IS_ACTIVE_FIELD_NUMBER: _ClassVar[int]
    id: str
    name: str
    content: str
    actions: str
    example_queries: ExampleQueries
    references: TopicReferences
    is_active: bool
    def __init__(self, id: _Optional[str] = ..., name: _Optional[str] = ..., content: _Optional[str] = ..., actions: _Optional[str] = ..., example_queries: _Optional[_Union[ExampleQueries, _Mapping]] = ..., references: _Optional[_Union[TopicReferences, _Mapping]] = ..., is_active: bool = ...) -> None: ...

class KnowledgeBase_DeleteTopic(_message.Message):
    __slots__ = ("id",)
    ID_FIELD_NUMBER: _ClassVar[int]
    id: str
    def __init__(self, id: _Optional[str] = ...) -> None: ...

class KnowledgeBase_ToggleTopicStatus(_message.Message):
    __slots__ = ("id", "is_active")
    ID_FIELD_NUMBER: _ClassVar[int]
    IS_ACTIVE_FIELD_NUMBER: _ClassVar[int]
    id: str
    is_active: bool
    def __init__(self, id: _Optional[str] = ..., is_active: bool = ...) -> None: ...

class KnowledgeBase_ImportTopics(_message.Message):
    __slots__ = ("created", "updated", "deleted", "source", "metadata")
    CREATED_FIELD_NUMBER: _ClassVar[int]
    UPDATED_FIELD_NUMBER: _ClassVar[int]
    DELETED_FIELD_NUMBER: _ClassVar[int]
    SOURCE_FIELD_NUMBER: _ClassVar[int]
    METADATA_FIELD_NUMBER: _ClassVar[int]
    created: _containers.RepeatedCompositeFieldContainer[KnowledgeBase_CreateTopic]
    updated: _containers.RepeatedCompositeFieldContainer[KnowledgeBase_UpdateTopic]
    deleted: _containers.RepeatedScalarFieldContainer[str]
    source: str
    metadata: ImportMetadata
    def __init__(self, created: _Optional[_Iterable[_Union[KnowledgeBase_CreateTopic, _Mapping]]] = ..., updated: _Optional[_Iterable[_Union[KnowledgeBase_UpdateTopic, _Mapping]]] = ..., deleted: _Optional[_Iterable[str]] = ..., source: _Optional[str] = ..., metadata: _Optional[_Union[ImportMetadata, _Mapping]] = ...) -> None: ...

class ImportMetadata(_message.Message):
    __slots__ = ("total_topics", "import_timestamp", "file_name")
    TOTAL_TOPICS_FIELD_NUMBER: _ClassVar[int]
    IMPORT_TIMESTAMP_FIELD_NUMBER: _ClassVar[int]
    FILE_NAME_FIELD_NUMBER: _ClassVar[int]
    total_topics: int
    import_timestamp: str
    file_name: str
    def __init__(self, total_topics: _Optional[int] = ..., import_timestamp: _Optional[str] = ..., file_name: _Optional[str] = ...) -> None: ...
